////////////////////////////////////////////////////////////////////// 
// merlin/Houdini.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Sunday March 07, 2004
// 
 
#include "Houdini.h"

#include <math.h>

double FuzzyInheritanceTree::perAlleleError = 0.0;
double FuzzyInheritanceTree::perGenotypeError = 0.0;
int    FuzzyInheritanceTree::errorCeiling = 0;

FuzzyInheritanceTree::FuzzyInheritanceTree()
   {
   stack_depth = 0;
   }

FuzzyInheritanceTree::~FuzzyInheritanceTree()
   {
   FreeStack();
   }

void FuzzyInheritanceTree::InitGenotypeProbs(Mantra & m)
   {
   // Allocate arrays for storing probability of observed genotypes
   // conditional on the number of matching alleles in the underlying
   // true genotype.
   match.Dimension(m.n);
   matchHi.Dimension(m.n);
   matchLo.Dimension(m.n);
   matchHomozygoteHi.Dimension(m.n);
   matchHomozygoteLo.Dimension(m.n);

   // And arrays for storing high and low alleles for each genotype
   hi.Dimension(m.two_n);
   lo.Dimension(m.two_n);

   // Since most possible underlying genotypes will differ from
   // the true genotype, we scale all probabilities to reduce
   // the number of multiplications.
   priorLikelihood = 1.0;

   // Calculate probability of each observed genotype conditional
   // on true genotype, using either a per allele or per genotype
   // error rate.
   if (perAlleleError > 0.0)
      AlleleErrorProbs(m);
   else
      GenotypeErrorProbs(m);
   }

void FuzzyInheritanceTree::GetAlleles(longint geno, int & allele1, int & allele2)
   {
   longint mask = 1;

   // Identify first allele
   allele1 = 1;
   while ((mask & geno) == 0)
      mask <<= 1, allele1++;

   // Identify second allele
   allele2 = allele1;
   if (mask != geno)
      do {
         mask <<= 1;
         allele2 ++;
      } while ((mask & geno) == 0);
   }

void FuzzyInheritanceTree::AlleleErrorProbs(Mantra & m)
   {
   // Error rate and its complement
   double & e = perAlleleError;
   double  ce = 1.0 - e;

   for (int i = 0; i < m.two_n; i+=2)
      if (m.genotype[i] != NOTZERO)
         {
         int allele1, allele2;

         GetAlleles(m.genotype[i], allele1, allele2);

         // Probability of genotype
         double base = m.frequencies[allele1] * m.frequencies[allele2] * e * e;

         if (allele1 != allele2) base *= 2;

         match[i >> 1] = (base + ((m.frequencies[allele1] + m.frequencies[allele2]) * e + ce) * ce) / base;

         matchLo[i >> 1] = (base + m.frequencies[allele2] * e * ce) / base;
         matchHomozygoteLo[i >> 1] = (base + 2 * m.frequencies[allele2] * e * ce) / base;
         matchHi[i >> 1] = (base + m.frequencies[allele1] * e * ce) / base;
         matchHomozygoteHi[i >> 1] = (base + 2 * m.frequencies[allele1] * e * ce) / base;

         lo[i] = lo[i + 1] = allele1;
         hi[i] = hi[i + 1] = allele2;

         priorLikelihood *= base;
         }
   }

void FuzzyInheritanceTree::GenotypeErrorProbs(Mantra & m)
   {
   // Error rate and its complement
   double & e = perGenotypeError;
   double  ce = 1.0 - e;

   for (int i = 0; i < m.two_n; i+=2)
      if (m.genotype[i] != NOTZERO)
         {
         int allele1, allele2;

         GetAlleles(m.genotype[i], allele1, allele2);

         // Probability of genotype
         double base = m.frequencies[allele1] * m.frequencies[allele2] * e;

         if (allele1 != allele2) base *= 2;

         match[i >> 1] = (base + ce) / base;

         lo[i] = lo[i + 1] = allele1;
         hi[i] = hi[i + 1] = allele2;

         priorLikelihood *= base;
         }

   matchLo.Set(1.0);
   matchHi = matchHomozygoteLo = matchHomozygoteHi = matchLo;
   }

void FuzzyInheritanceTree::FuzzyScoreVectors(Mantra & m, int mrk)
   {
   // If there is no genotyping error in the model, use faster
   // routines
   if (perAlleleError == 0.0 && perGenotypeError == 0.0)
      {
      ScoreVectors(m, mrk);
      return;
      }

   // Setup a framework for traversing inheritance vectors
   m.SelectMarker(mrk);

   // Setup probabilities for genotype matches and mismatches
   InitGenotypeProbs(m);

   // Clear the tree and define its depth
   Clear();
   bit_count = m.bit_count;

   // Setup up basic graphs and allele states, using founder genotypes
   AllocateStack(m);

   // The work horse
   ScoreRecursive(m, m.bit_count - 1, m.two_f, founderKeys, graphs,
                  graph_genotypes, m.two_f, m.two_f);

#ifndef __PRINTOUT_STATS__
   // Compact form for calculations
   if (aggressiveMerging) Trim(); else TrimNoMerge();
#endif
   }

int FuzzyInheritanceTree::ScoreRecursive
    (Mantra & m, int bit, int last_pivot,
    IntArray * founderKey, IntArray * graph, IntArray * graph_genos,
    int freec, int freef)
   {
   // Allocate a new tree node
   int node = NewNode();

   // Where in the inheritance vector are we now?
   int pivot = bit >= 0 ? m.bits[bit] : m.two_n;

   // Ensure graph consistency and track whether additional genotyped descendants
   // are possible for each graph
   graph_genos[1] = graph_genos[0];
   graph_genos++;

   for (int i = last_pivot; i < pivot; i++)
      {
      m.state[i] = m.state[m.vector[i]];

      // Adjust number of genotyped descendants, taking care to avoid twins
      if (m.isBit[i])
         // This line updates the number of potential genotyped descendants
         (*graph_genos)[(*graph)[m.state[m.vector[i] ^ 1]]] -= m.branch_genotyped[i];
      }

   // Update the graph of connected components in pedigree
   bool save_graph = false, save_key = false;
   for ( ; last_pivot < pivot; last_pivot++)
      {
      longint genotype = m.genotype[last_pivot];

      if (genotype == NOTZERO) continue;

      int founder = (*founderKey)[m.state[last_pivot]];

      if (alleleKey[founder][0] & genotype)
         {
         if (!save_key)
            {
            save_key = true;
            founderKey[1] = founderKey[0];
            founderKey++;
            }

         freqs[freef] = freqs[founder];
         alleleKey[freef] = alleleKey[founder];
         alleleIndex[freef] = alleleIndex[founder];

         if (alleleIndex[freef][hi[last_pivot]] == 0)
            ExposeAllele(m, freef, hi[last_pivot]);

         if (alleleIndex[freef][lo[last_pivot]] == 0)
            ExposeAllele(m, freef, lo[last_pivot]);

         (*founderKey)[m.state[last_pivot]] = freef;
         freef++;
         }

      if (last_pivot & 1)
         {
         int componentA = (*graph)[m.state[last_pivot - 1]];
         int componentB = (*graph)[m.state[last_pivot]];

         if (componentA != componentB)
            {
            if (!save_graph)
               {
               save_graph = true;
               graph[1] = graph[0];
               graph++;
               }

            (*graph_genos)[freec] = (*graph_genos)[componentA] +
                                    (*graph_genos)[componentB] - 2;

            components[freec] = components[componentA];
            components[freec].Append(components[componentB]);

            for (int i = 0; i < components[freec].Length(); i++)
               (*graph)[components[freec][i]] = freec;

            freec++;
            }
         else
            (*graph_genos)[componentA] -= 2;
         }
      }

   // Update likelihood components
   UpdateLikelihood(m, *founderKey, *graph, *graph_genos, last_pivot);

   // Is this a leaf node?
   if (bit < 0)
      {
      #ifdef __DEBUG_ERROR_GRAPH__
      printf("Genotypes: ");
      for (int i = 0; i < freec; i ++)
         printf("%3d ", (*graph_genos)[i]);
      printf("\n");
      #endif

//      for (int i = 0; i < m.two_f; i ++)
//         if (components[(*graph)[i]][0] == i)
//            if (((*graph_genos)[(*graph)[i]] != -1) || graph_likelihood[(*graph)[i]] <= 0.0)
//               printf("Trouble!");

      nodes[node].type = TREE_NODE_LEAF;
      nodes[node].value = priorLikelihood * CalculateLikelihood(m, *graph);

      return node;
      }

   // Find out if we can save time by using symmetries in the data
   switch (nodes[node].type = FindRedundancy(m, pivot))
      {
      case TREE_NODE_ZERO :
         nodes[node].value = 0.0;
         break;
      case TREE_NODE_ONE :
         m.state[pivot] = m.state[m.vector[pivot] &= ~1];
         SAFE_SET(nodes[node].child[0],
                  ScoreRecursive(m, bit - 1, last_pivot, founderKey, graph, graph_genos,
                                 freec, freef));
         break;
      case TREE_NODE_TWO :
         m.state[pivot] = m.state[m.vector[pivot] &= ~1];
         SAFE_SET(nodes[node].child[0],
                  ScoreRecursive(m, bit - 1, last_pivot, founderKey, graph, graph_genos,
                                 freec, freef));

         m.state[pivot] = m.state[m.vector[pivot] |= 1];
         SAFE_SET(nodes[node].child[1],
                  ScoreRecursive(m, bit - 1, last_pivot, founderKey, graph, graph_genos,
                                 freec, freef));
         break;
      }

   return node;
   }

int FuzzyInheritanceTree::FindRedundancy(Mantra & m, int pivot)
   {
   // Are there are any genotyped descendants?
   if (m.branch_genotyped[pivot] == 0)
      return TREE_NODE_ONE;

   // Are parents homozygous?
   int parent = m.vector[pivot];
   int matallele = m.state[parent & ~1];
   int fatallele = m.state[parent | 1];

   if (matallele == fatallele)
      return TREE_NODE_ONE;

   // If all the genotyped descendants of the two alternative
   // founder alleles are on this branch, it doesn't matter which
   // we pick -- one will go down, the other will be fixed (or not)
   if (m.branch_genotyped[matallele] == m.branch_genotyped[pivot] &&
       m.branch_genotyped[fatallele] == m.branch_genotyped[pivot] &&
       aggressiveMerging)
       return TREE_NODE_ONE;

   return TREE_NODE_TWO;
   }

double FuzzyInheritanceTree::CalculateLikelihood(Mantra & m, IntArray & graph)
   {
   double likelihood = 1.0;

   /* Go through each component */
   for (int i = 0; i < m.two_f; i ++)
      if (components[graph[i]][0] == i)
         likelihood *= graph_likelihood[graph[i]];

   return likelihood;
   }

void FuzzyInheritanceTree::UpdateLikelihood(Mantra & m, IntArray & founderKey,
      IntArray & graph, IntArray & graph_genos, int pivot)
   {
   for (int i = 0; i < m.two_f; i ++)
      {
      int c = graph[i];

      if ((graph_genos[c] == 0) && (m.vector[i] == -1))
         {
         graph_genos[c] = -1;

         PrepareComponent(m, graph, founderKey, c, pivot);
         CollapsePairs(m, graph, c);
         graph_likelihood[c] = ComponentLikelihood(m, graph, c);

         #ifdef __DEBUG_ERROR_GRAPH__
         DebugGraph(m, founderKey, graph, c, i);
         #endif
         }
      }
   }

double FuzzyInheritanceTree::ComponentLikelihood
   (Mantra & m, IntArray & graph, int component, longint * bestAlleleStates)
   {
   IntArray & mapped = mappedComponent;
   IntArray & original = activeComponent;

   int alleles = original.Length();

   // Find the first unique allele for each descent line
   indexBase.Dimension(alleles);
   for (int i = 0; i < alleles; i++)
      indexBase[i] = alleleKey[mapped[i]][0] ? 0 : 1;
   index = indexBase;

   // Initialize allele states
   allele.Dimension(m.two_n);
   for (int i = 0; i < alleles; i++)
      allele[original[i]] = alleleKey[mapped[i]][index[i]];

   double lk = 0.0, best = 0.0, partial;

   while (true)
      {
      lk += partial = PartialLikelihood(m, graph, component);

      if ((bestAlleleStates != NULL) && partial > best)
         {
         for (int i = 0; i < alleles; i++)
            bestAlleleStates[original[i]] = index[i] ? (allele[original[i]]) : (NOTZERO);

         best = partial;
         }

      int i = 0;
      while (i < alleles && index[i] == (alleleKey[mapped[i]].Length() - 1))
         i++;

      if (i == alleles)
         break;

      for (int j = 0; j < i; j++)
         {
         index[j]  = indexBase[j];
         allele[original[j]] = alleleKey[mapped[j]][indexBase[j]];
         }

      allele[original[i]] = alleleKey[mapped[i]][++index[i]];
      }

   return lk;
   }

double FuzzyInheritanceTree::PartialLikelihood(Mantra & m, IntArray & graph, int component)
   {
   double  lk = 1.0;
   int errors = errorCeiling - 1;

   // Prior probability of founder allele states
   for (int i = 0; i < activeComponent.Length(); i++)
      lk *= (*(cFreqMap[activeComponent[i]]))[index[i]];

   // Probability of observed genotypes conditional on founder allele states
   for (int j = 0; j < cGeno.Length(); j++)
      {
      int pheno = cGeno[j];
      int al1 = allele[cState[j * 2]];
      int al2 = allele[cState[j * 2 + 1]];

      int i = cMap[j];

      if (pheno == (al1 | al2))
         lk *= match[i];
      else if ((pheno & al1) & (pheno & al2))
         lk *= (pheno & al1) > (pheno ^ al1) ? matchHomozygoteHi[i] : matchHomozygoteLo[i];
      else if (pheno & al1)
         lk *= (pheno & al1) > (pheno ^ al1) ? matchHi[i] : matchLo[i];
      else if (pheno & al2)
         lk *= (pheno & al2) > (pheno ^ al2) ? matchHi[i] : matchLo[i];
      else if (errors-- == 0)
         return 0.0;
      }

   return lk;
   }

void FuzzyInheritanceTree::PrepareComponent(Mantra & m, IntArray & graph,
      IntArray & founderKey, int c, int pivot)
   {
   // Setup temporary list of component descent graphs
   activeComponent = components[c];
   mappedComponent.Dimension(components[c].Length());
   for (int j = 0; j < activeComponent.Length(); j++)
      mappedComponent[j] = founderKey[activeComponent[j]];

   // List genotyped descendants
   cGeno.Clear(); cState.Clear(); cMap.Clear();
   for (int i = 0; i < pivot; i += 2)
      if ((m.genotype[i] != NOTZERO) && (graph[m.state[i]] == c))
         {
         cGeno.Push(m.genotype[i]);
         cState.Push(m.state[i]);
         cState.Push(m.state[i+1]);
         cMap.Push(i >> 1);
         }

   // List component allele frequencies
   for (int j = 0; j < mappedComponent.Length(); j++)
      cFreqMap[activeComponent[j]] = &(freqs[mappedComponent[j]]);
   }

void FuzzyInheritanceTree::CollapsePairs(Mantra & m, IntArray & graph, int component)
   {
   int next_freq = 0;

   while (activeComponent.Length() > 2)
      {
      allelePairs.Dimension(m.two_f);
      allelePairs.Set(-1);

      // Identify pairs of founder alleles
      for (int i = 0; i < cState.Length(); i++)
         if (allelePairs[cState[i]] == -1)
            allelePairs[cState[i]] = cState[i ^ 1];
         else if (allelePairs[cState[i]] != cState[i ^ 1])
            allelePairs[cState[i]] = -2;

      // Check if any alleles are suitable for collapsing
      int success = false;
      for (int i = 0; i < activeComponent.Length(); i++)
         if ((allelePairs[activeComponent[i]] >= 0) &&
             (allelePairs[activeComponent[i]] != activeComponent[i]))
            {
            int collapse = activeComponent[i];
            int pair = allelePairs[collapse];
            int pairi = activeComponent.Find(pair);

            // Extract genotypes for this pair from master list
            pGeno.Clear(); pMap.Clear();
            for (int j = cGeno.Length() - 1; j >= 0; j--)
               if (cState[j * 2] == collapse || cState[j * 2 + 1] == collapse)
                  {
                  pGeno.Push(cGeno[j]);
                  pMap.Push(cMap[j]);
                  cGeno.Delete(j);
                  cMap.Delete(j);
                  cState.Delete(j * 2);
                  cState.Delete(j * 2);
                  }

            cFreqs[next_freq] = *(cFreqMap[pair]);
            cFreqMap[pair] = &(cFreqs[next_freq]);
            next_freq++;

            IntArray & map1 = alleleKey[mappedComponent[pairi]];
            IntArray & map2 = alleleKey[mappedComponent[i]];

            for (int state1 = 0; state1 < map1.Length(); state1++)
               if (map1[state1])
               {
               int al1 = map1[state1];
               double factor = 0.0;

               for (int state2 = 0; state2 < map2.Length(); state2++)
                  if (map2[state2])
                     {
                     int al2 = map2[state2];
                     double term = (*(cFreqMap[collapse]))[state2];

                     for (int k = 0; k < pGeno.Length(); k++)
                        {
                        int pheno = pGeno[k];
                        int l = pMap[k];

                        if (pheno == (al1 | al2))
                           term *= match[l];
                        else if ((pheno & al1) & (pheno & al2))
                           term *= (pheno & al1) > (pheno ^ al1) ? matchHomozygoteHi[l] : matchHomozygoteLo[l];
                        else if (pheno & al1)
                           term *= (pheno & al1) > (pheno ^ al1) ? matchHi[l] : matchLo[l];
                        else if (pheno & al2)
                           term *= (pheno & al2) > (pheno ^ al2) ? matchHi[l] : matchLo[l];
                        }

                     factor += term;
                     }

               (*(cFreqMap[pair]))[state1] *= factor;
               }

            activeComponent.Delete(i);
            mappedComponent.Delete(i);

            success = true;

            break;
            }

      if (!success) return;
      }
   }

void FuzzyInheritanceTree::AllocateStack(Mantra & m)
   {
   // Work out a reasonable stack depth
   int new_depth = m.bit_count;

   if (m.two_n > new_depth)
      new_depth = m.two_n;

   // new_depth = (new_depth & ~31) + 32;

   // Allocate stack if necessary
   if (new_depth > stack_depth)
      {
      FreeStack();

      stack_depth = new_depth;

      components = new IntArray [stack_depth];
      graphs = new IntArray [stack_depth];
      graph_genotypes = new IntArray [stack_depth];
      founderKeys = new IntArray [stack_depth];
      alleleIndex = new IntArray [stack_depth];
      alleleKey = new IntArray [stack_depth];
      freqs = new Vector[stack_depth];
      cFreqs = new Vector[stack_depth];
      cFreqMap = new Vector * [stack_depth];

      graph_likelihood.Dimension(stack_depth);
      }

   // Initialize graph
   graphs[0].Dimension(m.two_f);
   graphs[0].SetSequence(0, 1);

   // Initialize founder allele keys
   founderKeys[0].Dimension(m.two_n);
   founderKeys[0].SetSequence(0, 1);

   // Initialize components for each singleton graph
   for (int i = 0; i < m.two_f; i++)
      {
      components[i].Dimension(1);
      components[i][0] = m.vector[i] >= 0 ? -1 : i;
      }

   int alleleCount = m.frequencies.Length();
   int alleleMask = (((1 << (alleleCount - 2)) - 1) << 1) + 1;

   // Initialize allele sets for each component
   for (int i = 0; i < m.two_f; i++)
      {
      alleleIndex[i].Dimension(alleleCount);
      alleleIndex[i].Set(0);

      alleleKey[i].Dimension(1);
      alleleKey[i][0] = alleleMask;

      freqs[i].Dimension(1);
      freqs[i][0] = 1.0;
      }

   graph_genotypes[0] = m.branch_genotyped;

   // Setup graphs and founder components for genotyped founders
   for (int i = 0; i < m.two_f; i+=2)
      if (m.genotype[i] != NOTZERO)
         {
         if (m.vector[i + 1] == -1)
            {
            // This code is executed for most founders
            graphs[0][i + 1] = graphs[0][i];
            components[i].Append(components[i + 1]);
            graph_genotypes[0][i] += graph_genotypes[0][i + 1] - 2;
            }
         else
            // This code is executed for males when analysing
            // the X chromosome. In this circunstance, there
            // is only one founder allele to consider, and we
            // don't need to merge two graphs
            graph_genotypes[0][i] -= 2;

         ExposeAllele(m, i, lo[i]);

         if (m.hetero[i])
            ExposeAllele(m, i, hi[i]);

         if (m.vector[i + 1] == -1)
            {
            alleleIndex[i + 1] = alleleIndex[i];
            alleleKey[i + 1] = alleleKey[i];
            freqs[i + 1] = freqs[i];
            }
         }
   }

void FuzzyInheritanceTree::FreeStack()
   {
   if (stack_depth)
      {
      // Delete temporary variables from stack
      delete [] components;
      delete [] graphs;
      delete [] graph_genotypes;
      delete [] founderKeys;
      delete [] alleleIndex;
      delete [] alleleKey;
      delete [] freqs;
      delete [] cFreqs;
      delete [] cFreqMap;
      }
   }

void FuzzyInheritanceTree::DebugGraph(Mantra & m,
   IntArray & founderKey, IntArray & graph, int c, int i)
   {
   printf("\nCOMPONENT %d", i);
   for (int j = 0; j < components[c].Length(); j++)
      {
      printf("\n  Founder Allele %c", 'A' + components[c][j]);

      int map = founderKey[components[c][j]];
      printf("\n    Index: ");
      for (int k = 0; k < alleleIndex[map].Length(); k++)
         if (alleleIndex[map][k] || k == 0)
            printf(" %5d ", k);

      printf("\n    Freqs: ");
      for (int k = 0; k < alleleIndex[map].Length(); k++)
         if (alleleIndex[map][k] || k == 0)
            printf(" %5.3f ", freqs[map][alleleIndex[map][k]]);

      printf("\n    Masks: ");
      for (int k = 0; k < alleleIndex[map].Length(); k++)
         if (alleleIndex[map][k] || k == 0)
            printf(" %5d ", alleleKey[map][alleleIndex[map][k]]);
      }
   printf("\n  Genotypes: ");
   for (int k = 0; k < m.two_n; k += 2)
      if ((m.genotype[k] != NOTZERO) && (graph[m.state[k]] == graph[i]))
         printf(" %c/%c = %d/%d; ",
                'A' + m.state[k], 'A' + m.state[k + 1],
                hi[k], lo[k]);

   int last = 0;
   for (int k = 0; k < m.two_n; k += 2)
      if ((m.genotype[k] != NOTZERO) && (graph[m.state[k]] == graph[i]))
         last = k;
   int savings = m.bits.CountIfGreater(last);
   printf("\n  Final Meiosis: %d / %d; Savings: %d\n", last, m.two_n, savings);

   printf("\n  Likelihood = %5.3f\n", ComponentLikelihood(m, graph, graph[i]));
   }

double FuzzyInheritanceTree::FuzzyEvaluateInheritanceVector(Mantra & m, int * vector,
       int * g, longint * alleles, longint * alleles2, int * fixed)
   {
   // If there is no genotyping error in the model, use faster routines
   if (perAlleleError == 0.0 && perGenotypeError == 0.0)
      return EvaluateInheritanceVector(m, vector, g, alleles, alleles2, fixed);

   // Initialize essential data structures
   InitGenotypeProbs(m);

   // Setup up basic graphs and allele states, using founder genotypes
   AllocateStack(m);

   // Setup the inheritance vector
   for (int i = 0; i < m.bit_count; i++)
      {
      int bit = m.bits[i];

      m.vector[bit] = vector[i] == 0 ? m.vector[bit] & ~1 : m.vector[bit] | 1;
      }

   // Identify founder alleles for each individual
   for (int i = m.two_f; i < m.two_n; i++)
      m.state[i] = m.state[m.vector[i]];

   for (int i = m.two_f; i < m.two_n; i++)
      m.state[i] = m.state[m.vector[i]];

   IntArray & founderKey = founderKeys[0];
   IntArray & graph = graphs[0];
   IntArray & graph_genos = graph_genotypes[0];

   // Update the graph of connected components in pedigree
   for ( int i = m.two_f; i < m.two_n; i++)
      {
      longint genotype = m.genotype[i];

      if (genotype == NOTZERO) continue;

      int founder = founderKey[m.state[i]];

      if (alleleKey[founder][0] & genotype)
         {
         if (alleleIndex[founder][hi[i]] == 0)
            ExposeAllele(m, founder, hi[i]);

         if (alleleIndex[founder][lo[i]] == 0)
            ExposeAllele(m, founder, lo[i]);
         }

      if (i & 1)
         {
         int componentA = graph[m.state[i - 1]];
         int componentB = graph[m.state[i]];

         if (componentA != componentB)
            {
            components[componentA].Append(components[componentB]);

            for (int j = 0; j < components[componentB].Length(); j++)
               graph[components[componentB][j]] = componentA;
            }
         }
      }

   for (int i = 0; i < m.two_f; i++)
      if (graph[i] == i)
         graph_genos[i] = 0;

   // Update likelihood components
   FillAlleleStates(m, founderKey, graph, graph_genos, alleles);

   for (int i = 0; i < m.two_f; i++)
      {
      fixed[i] = alleles[i] != NOTZERO;
      g[i] = i;
      }

   return priorLikelihood * CalculateLikelihood(m, graph);
   }

void FuzzyInheritanceTree::FillAlleleStates(Mantra & m, IntArray & founderKey,
      IntArray & graph, IntArray & graph_genos, longint * alleles)
   {
   for (int i = 0; i < m.two_f; i ++)
      {
      int c = graph[i];

      if ((graph_genos[c] == 0) && (m.vector[i] == -1))
         {
         graph_genos[c] = -1;

         PrepareComponent(m, graph, founderKey, c, m.two_n);
         graph_likelihood[c] = ComponentLikelihood(m, graph, c, alleles);
         }
      }
   }

 
