////////////////////////////////////////////////////////////////////// 
// merlin/Houdini.h 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Sunday March 07, 2004
// 
 
#ifndef __HOUDINI_H__
#define __HOUDINI_H__

#include "Magic.h"

class FuzzyInheritanceTree : public InheritanceTree
   {
   public:
      FuzzyInheritanceTree();
      ~FuzzyInheritanceTree();

      // Scores the likelihoods of all inheritance vectors
      void FuzzyScoreVectors(Mantra & m, int marker);

      // Scores the likelihood of a single inheritance vector
      double  FuzzyEvaluateInheritanceVector(Mantra & m, int * vector,
              int * graph, longint * alleles, longint * alleles2, int * fixed);

      // The probability that an allele is called incorrectly
      static double perAlleleError;
      static double perGenotypeError;
      static int    errorCeiling;

   private:
      // Stacks for the most commonly used arrays
      int stack_depth;

      // Routines for managing temporary storage
      void AllocateStack(Mantra & m);
      void FreeStack();

      // Recurse throught the pedigree and store the likelihood of all
      // possible inheritance vectors
      int ScoreRecursive(Mantra & m, int current_bit, int last_pivot,
               IntArray * founderKey, IntArray * graph, IntArray * graph_genos,
               int freec, int freef);

      // Check if branching is required (while recursing through the pedigree)
      int FindRedundancy(Mantra & m, int pivot);

      // Calculate the likelihood for a single inheritance vector
      void   UpdateLikelihood(Mantra & m, IntArray & founderKey, IntArray & graph, IntArray & graph_genos, int pivot);
      void   FillAlleleStates(Mantra & m, IntArray & founderKey, IntArray & graph, IntArray & graph_genos, longint * alleles);
      double CalculateLikelihood(Mantra & m, IntArray & graph);

      // Calculate likelihood for a component of the inheritance graph
      void   PrepareComponent(Mantra & m, IntArray & graph, IntArray & founderKey, int component, int pivot);
      double ComponentLikelihood(Mantra & m, IntArray & graph, int component, longint * bestAlleleStates = NULL);
      double PartialLikelihood(Mantra & m, IntArray & graph, int component);

      // Collapse founder allele pairs, to speed up ComponentLikelihood
      void   CollapsePairs(Mantra & m, IntArray & graph, int component);

      // Calculate probabilities of matches and mismatches for
      // each genotype
      void InitGenotypeProbs(Mantra & m);
      void AlleleErrorProbs(Mantra & m);
      void GenotypeErrorProbs(Mantra & m);
      void GetAlleles(longint genotype, int & allele1, int & allele2);

      // Expose an allele for individual tracking in a descent line
      void ExposeAllele(Mantra & m, int line, int allele)
         {
         int alias = alleleKey[line].Length();
         int  mask = 1 << (allele - 1);
         double  f = m.frequencies[allele];

         alleleIndex[line][allele] = alias;
         alleleKey[line].Push(mask);
         alleleKey[line][0] ^= mask;
         freqs[line].Push(f);
         freqs[line][0] -= f;
         }

      // Print debugging information for each graph
      void DebugGraph(Mantra & m, IntArray & founderKey, IntArray & graph, int c, int i);

      // Likelihood of an inheritance vector independent of marker data
      double priorLikelihood;

      // Probabilities of observed genotypes as a function of underlying
      // true genotype
      Vector match, matchHi, matchLo, matchHomozygoteHi, matchHomozygoteLo;

      // Allelic state information for each component
      IntArray allele, indexBase, index;
      IntArray activeComponent, mappedComponent;

      // Allelic components for each genotype
      IntArray hi, lo;

      // Connected components in founder allele graph
      IntArray * components;
      IntArray * graphs;

      // Structures to allow early calculation of graph components
      IntArray * graph_genotypes;
      Vector     graph_likelihood;

      // Structures for tracking distinct alleles for each founder
      IntArray * founderKeys;
      IntArray * alleleIndex;
      IntArray * alleleKey;
      Vector   * freqs;

      // Temporary storage for ComponentLikelihood and PartialLikelihood
      IntArray cGeno, cState, cMap;
      Vector * cFreqs, ** cFreqMap;

      // Temporary storage for Pair Finding and Collapsing routines
      IntArray pGeno, pMap;
      IntArray allelePairs;
   };


#endif
 
