////////////////////////////////////////////////////////////////////// 
// merlin/MerlinCore.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Sunday March 07, 2004
// 
 
#include "MerlinCore.h"
#include "MapFunction.h"
#include "Houdini.h"
#include "Error.h"

#include <math.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>

// Limits to stop Merlin from attempting problems that are too tough
int MerlinCore::maxBits = 24;
int MerlinCore::maxMinutes = 0;

// Positions to analyse
int    MerlinCore::steps_per_interval = 0;
double MerlinCore::maxDistance = _NAN_;
double MerlinCore::minDistance = _NAN_;
double MerlinCore::grid  = _NAN_;
double MerlinCore::start = _NAN_;
double MerlinCore::stop  = _NAN_;

// Output options
bool MerlinCore::quietOutput = false;
bool MerlinCore::useMarkerNames = false;
String MerlinCore::filePrefix = "merlin";

// Swap options
bool MerlinCore::useSwap = false;

// Internal flags to minimize useless calculations
bool MerlinCore::multipoint = false;
bool MerlinCore::scanChromosome = false;
bool MerlinCore::betweenMarkers = false;

// Flags for controlling recombination
bool MerlinCore::zeroRecombination = false;
bool MerlinCore::oneRecombination = false;
bool MerlinCore::twoRecombinations = false;
bool MerlinCore::threeRecombinations = false;
bool MerlinCore::twopoint = false;

// Error model
double MerlinCore::imaginaryTheta = 0.00;

// Determines whether error checking stub should be executed
bool MerlinCore::findErrors = false;

// Constructor and Destructor
//

MerlinCore::MerlinCore(Pedigree & pedigree) : mantra(), ped(pedigree)
   {
   printHeader = false;
   lowBound = pow(2.0, -129);
   rescale  = pow(2.0, 258);
   lnScale  = 258 * log(0.5);
   }

MerlinCore::~MerlinCore()
   {
   }

// Global setup that must be carried out before calculating likelihods
//

void MerlinCore::SetupGlobals()
   {
   // Set the maximum number of recombinants in multipoint calculations
   Multipoint::maximum_recombinants =
      threeRecombinations * 3 + twoRecombinations * 2 + oneRecombination;
   }

// Setup map for analysing the next chromosome at currently selected locations
//

int MerlinCore::SetupMap(int chromosome)
   {
   markers.Dimension(0);
   int next = PedigreeGlobals::SortMarkersInMapOrder(markers, chromosome);

   markerCount = markers.Length();

   if (markerCount == 0) error("No markers could be placed");

   // Check that stop position for analysis is greater than start position
   if (MerlinCore::start != _NAN_ && MerlinCore::stop != _NAN_ &&
       MerlinCore::start > MerlinCore::stop)
      warning("Requested position for starting analyses is greater than final position\n");

   // Clear previous information
   labels.Clear();
   analysisPositions.Clear();

   // Allocate arrays of known size
   markerPositions.Dimension(markerCount);
   markerFemalePositions.Dimension(markerCount);
   markerMalePositions.Dimension(markerCount);

   String label;

   // Retrieve marker positions
   for (int i = 0; i < markerCount; i++)
      {
      MarkerInfo * info = PedigreeGlobals::GetMarkerInfo(markers[i]);
      ValidateMarker(info);
      markerPositions[i] = info->position;
      markerMalePositions[i] = info->positionMale;
      markerFemalePositions[i] = info->positionFemale;
      }

   // For maps with no recombination, only analyse one location
   if (zeroRecombination)
      {
      analysisPositions.Push(markerPositions[0]);
      labels.Push(useMarkerNames ?
                  PedigreeGlobals::GetMarkerInfo(markers[0])->name :
                  label = (markerPositions[0] * 100.0));
      }
   // Singlepoint analysis at marker locations only
   else if (twopoint)
      {
      analysisPositions = markerPositions;

      for (int i = 0; i < markerCount; i++)
         labels.Push(PedigreeGlobals::GetMarkerInfo(markers[i])->name);
      }
   // Analysis along grid of equally spaced locations
   else if (grid != _NAN_)
      {
      double position = start != _NAN_ ? start * 0.01 :
         PedigreeGlobals::GetMarkerInfo(markers[0])->position;
      double mapEnd = stop != _NAN_ ? stop * 0.01 :
         PedigreeGlobals::GetMarkerInfo(markers.Last())->position;
      double step = grid * 0.01;

      while (true)
         {
         analysisPositions.Push(position);
         labels.Push(label = position * 100.0);
         if (mapEnd - position < 1e-5) break;
         position += step;
         }
      }
   // Analysis at marker positions and equally spaced locations in-between
   else
      {
      double min = minDistance == _NAN_ ? _NAN_ : minDistance * 0.01;
      double max = maxDistance == _NAN_ ? _NAN_ : maxDistance * 0.01;

      double last = markerPositions[0];

      for (int i = 0; i < markerCount; i++)
         {
         if (markerPositions[i] > last)
            {
            double delta = markerPositions[i] - last;
            int    steps = steps_per_interval;
            double step  = delta / (steps + 1);

            if ((max != _NAN_) && (max < step))
               {
               steps = (int) (delta / (max + 1e-6));
               step  = delta / (steps + 1);
               }

            if ((min != _NAN_) && min > step)
               steps = (int) (delta / (minDistance  + 1e-6) - 0.999);

            if (steps < 0) steps = 0;

            for (int j = 1; j <= steps; j++)
               {
               double position = last + delta * j / (steps+1);

               analysisPositions.Push(position);
               labels.Push(label = position * 100.0);
               }
            }

         if (markerPositions[i] > last || i == 0)
            {
            analysisPositions.Push(last = markerPositions[i]);
            labels.Push(useMarkerNames ?
               PedigreeGlobals::GetMarkerInfo(markers[i])->name :
               label = (last * 100.0));
            }
         }

      if (start != _NAN_)
         {
         int firstPosition = analysisPositions.Length() -
                             analysisPositions.CountIfGreaterOrEqual(start * 0.01);

         if (firstPosition > 0)
            {
            for (int i = firstPosition; i < analysisPositions.Length(); i++)
               {
               analysisPositions[i - firstPosition] = analysisPositions[i];
               labels[i - firstPosition] = labels[i];
               }

            analysisPositions.Dimension(analysisPositions.Length() - firstPosition);
            labels.Dimension(analysisPositions.Length());
            }
         }

      if (stop != _NAN_)
         {
         int lastPosition = analysisPositions.Length() -
                            analysisPositions.CountIfGreater(stop * 0.01);

         analysisPositions.Dimension(lastPosition);
         labels.Dimension(lastPosition);
         }
      }

   SexSpecificMap();

   return next;
   }

// Construct sex specific map using current sex averaged map as a template
//

void MerlinCore::SexSpecificMap()
   {
   // First calculate recombination fractions between consecutive markers
   femaleMarkerTheta.Dimension(markerCount - 1);
   maleMarkerTheta.Dimension(markerCount - 1);
   keyFemaleTheta.Dimension(markerCount - 1);
   keyMaleTheta.Dimension(markerCount - 1);

   for (int i = 1; i < markerCount; i++)
      {
      MarkerInfo * info1 = ped.GetMarkerInfo(markers[i - 1]);
      MarkerInfo * info2 = ped.GetMarkerInfo(markers[i]);

      femaleMarkerTheta[i - 1] =
         DistanceToRecombination(info2->positionFemale - info1->positionFemale);
      maleMarkerTheta[i - 1] =
         DistanceToRecombination(info2->positionMale - info1->positionMale);
      }

   // Check if there are any analysis positions ...
   if (analysisPositions.Length() == 0) return;

   // Pre-allocate arrays to speed things up
   malePositions.Dimension(analysisPositions.Length());
   femalePositions.Dimension(analysisPositions.Length());

   // Variables to keep track of current position
   int pos = 0;
   int marker = 0;

   // Next calculate recombination fractions between each analysis position an
   // the first marker (using the sex averaged map)
   while (analysisPositions[pos] < markerPositions[marker])
      {
      double delta = analysisPositions[pos] - markerPositions[marker];
      malePositions[pos] = markerMalePositions[marker] + delta;
      femalePositions[pos] = markerFemalePositions[marker] + delta;
      pos++;
      }

   // Calculate recombination fractions for all intervening positions
   while (marker + 1 < markerCount)
      {
      marker++;

      double delta = markerPositions[marker] - markerPositions[marker - 1];
      double maleDelta = markerMalePositions[marker] - markerMalePositions[marker - 1];
      double femaleDelta = markerFemalePositions[marker] - markerFemalePositions[marker - 1];

      while (pos < analysisPositions.Length() &&
             analysisPositions[pos] < markerPositions[marker])
         {
         double f = (analysisPositions[pos] - markerPositions[marker - 1]) / delta;

         malePositions[pos] = markerMalePositions[marker - 1] + maleDelta * f;
         femalePositions[pos] = markerFemalePositions[marker - 1] + femaleDelta * f;
         pos++;
         }
      }

   // Calculate recombination fractions between final analysis positions and
   // the last marker (using the sex averaged map)
   while (analysisPositions.Length() > pos)
      {
      double delta = analysisPositions[pos] - markerPositions[marker];
      malePositions[pos] = markerMalePositions[marker] + delta;
      femalePositions[pos] = markerFemalePositions[marker] + delta;
      pos++;
      }
   }

// Setup analyses for a new family
//

bool MerlinCore::SelectFamily(Family * f, bool warnOnSkip)
   {
   mantra.Prepare(ped, *f);
   mantra.PrepareIBD();

   printHeader = true;
   cleanOutput = false;

   family = f;
   serial = f->serial;

   // if (mantra.couples)
   //   PrintMessage("  %d symmetric founder couple%s detected",
   //               mantra.couples, mantra.couples == 1 ? "" : "s");

   if (mantra.bit_count <= maxBits)
      return true;

   if (warnOnSkip)
      PrintMessage("  SKIPPED: Maximum complexity set at %d bits", maxBits);
   CleanMessages();
   return false;
   }

// Calculate Likelihood of Observed Marker Data
//

double MerlinCore::CalculateLikelihood()
   {
   singlepoint = NULL;
   right = NULL;

   try {
      BasicTree::FreeSwap();

      ScoreSinglepoint();

      if (informativeCount > 0)
         ScoreConditionals();

      CleanMessages();

      if (right != NULL)
         delete [] right;
      delete [] singlepoint;
   }
   catch (const TreesTooBig & problem)
      {
      PrintMessage("  SKIPPED: At least %d megabytes needed", problem.memory_request);
      CleanMessages();

      if (singlepoint != NULL) delete [] singlepoint;
      if (right != NULL) delete [] right;
      }
   catch (const OutOfTime & problem)
      {
      PrintMessage("  SKIPPED: Analysis would require more than %d minutes\n", maxMinutes);
      CleanMessages();

      if (singlepoint != NULL) delete [] singlepoint;
      if (right != NULL) delete [] right;
      };

   return likelihood;
   }

// Calculation of single marker likelihoods
//

void MerlinCore::ScoreSinglepoint()
   {
   keyThetas = false;

   singlepoint = new InheritanceTree[ped.markerCount];
   FuzzyInheritanceTree tempVectors;

   informativeCount = 0;
   informativeMarkers.Clear();
   informativePositions.Clear();
   informationScores.Clear();
   informationBounds.Clear();

#ifdef __PRINTOUT_STATS__
   Vector nodeCounts, leafCounts, zeroCounts, infoScores;
#endif

   // Reset the likelihood
   likelihood = 0;

   for (int m = 0; m < markerCount; m++)
      {
      ProgressReport("Singlepoint", m, markerCount);
      tempVectors.FuzzyScoreVectors(mantra, markers[m]);

#ifndef __PRINTOUT_STATS__
      stats.GetTreeInfo(tempVectors);
#else
      stats.GetDetailedInfo(tempVectors);
      nodeCounts.Push(stats.nodeCount);
      leafCounts.Push(stats.leafCount);
      zeroCounts.Push(stats.zeroCount);
      infoScores.Push(stats.information);
#endif

      if (stats.mean == 0.0)
         {
         // TODO -- be helpful and try to pin down problem genotypes
         PrintMessage("  Skipping Marker %-10s [BAD INHERITANCE]",
              (const char *) ped.markerNames[markers[m]]);

         // Replace with uninformative inheritance tree
         tempVectors.MakeMinimalTree(1.0, mantra.bit_count);
         stats.information = 0.0;

         // Erase genotypes for this marker
         for (int i = family->first; i <= family->last; i++)
            ped[i].markers[markers[m]][2] = ped[i].markers[markers[m]][1] = 0;
         }

      if (stats.information > 1e-5 || twopoint ||
           InheritanceTree::aggressiveMerging == false &&
           tempVectors.nodes[0].type != TREE_NODE_ZERO &&
           tempVectors.nodes[0].type != TREE_NODE_LEAF)
         {
         // Comenting out the next few lines enables
         // the complex recombination fraction of Terwilliger and Goring
         //

         if (imaginaryTheta)
            {
            Multipoint projection;

            projection.Copy(tempVectors);
            projection.MoveAlong(mantra, imaginaryTheta, 1);
            tempVectors.Copy(projection);
            }

         singlepoint[informativeCount].Copy(tempVectors);
         singlepoint[informativeCount++].Pack();
         informativeMarkers.Push(m);
         informativePositions.Push(markerPositions[m]);
         informationScores.Push(stats.information);
         informationBounds.Push(stats.information > 0.3 ? stats.GetMinInfo(tempVectors) : 0.0);
         }
      else
         {
         // Likelihood for all gene flow patterns is effectively equal
         // if (stats.information > 1e-10 && !quietOutput)
         //   PrintMessage("  Skipping Marker %-10s [Info = %.6f]",
         //       (const char *) ped.markerNames[markers[m]], stats.information);

         // stats.mean as calculated by GetTreeInfo must be scaled
         if (stats.mean)
            likelihood += log(stats.mean) + mantra.bit_count * log(0.5);
         }
      }

   if (Multipoint::maximum_recombinants || twopoint || zeroRecombination ||
       mantra.bit_count < 3)
      useSparse = false;
   else
      {
      // By default we don't assume sparseness
      useSparse = false;

      // Lower information bounds if founder couple symmetries are being used
      if (mantra.couples)
         // TODO -- work out what a better penalty for couples is ...
         informationBounds.Add(-1.0 * mantra.couples / mantra.bit_count);

      // Unless very informative marker pairs are present
      for (int i = 0; i < informativeCount - 1; i++)
         if (informationBounds[i] + informationBounds[i + 1] > 1.0)
            useSparse = true;
      }

#ifdef __PRINTOUT_STATS__
   int median = nodeCounts.dim / 2;
   int ciLo = nodeCounts.dim * 25 / 1000;
   int ciHi = nodeCounts.dim * 975 / 1000;

   char * labels [] = { "Nodes", "Leafs", "Zeros", "Info" };
   Vector * vect [] = { &nodeCounts, &leafCounts, &zeroCounts, &infoScores};

   for (int i = 0; i < 4; i++)
      {
      Vector & vector = *vect[i];

      vector.Sort();
      PrintMessage("  % 8s: %.2f (%.2f to %.2f) Median = %.2f, 95%% CI = %.2f - %.2f",
         labels[i], vector.Average(), vector.Min(), vector.Max(),
         vector[median], vector[ciLo], vector[ciHi]);
      }
#endif
   }

// Calculation of between marker recombination fractions
//

void MerlinCore::CalculateThetas()
   {
   if (keyThetas) return;

   // keyFemaleTheta.Dimension(informativeCount - 1);
   // keyMaleTheta.Dimension(informativeCount - 1);

   for (int m = informativeCount - 2; m >= 0; m--)
      {
      double theta[2];
      int    mrk = informativeMarkers[m];

      theta[0] = 1.0 - 2.0 * femaleMarkerTheta[mrk];
      theta[1] = 1.0 - 2.0 * maleMarkerTheta[mrk];

      while (++mrk < informativeMarkers[m + 1])
         {
         theta[0] *= 1.0 - 2.0 * femaleMarkerTheta[mrk];
         theta[1] *= 1.0 - 2.0 * maleMarkerTheta[mrk];
         }

      theta[0] = (1.0 - theta[0]) / 2;
      theta[1] = (1.0 - theta[1]) / 2;

      keyFemaleTheta[m] = theta[0];
      keyMaleTheta[m] = theta[1];
      }

   keyThetas = true;
   }

// Calculation of right conditional likelihoods
//

bool MerlinCore::ScoreConditionals()
   {
   if (zeroRecombination)
      return ZeroRecombination();

   if (twopoint) return true;

   CalculateThetas();

   right = new Tree[informativeCount];
   rightScale.Dimension(informativeCount);
   leftScale.Dimension(informativeCount);

   Multipoint engine;

   if (useSparse) engine.SetupConditioning(mantra);

   int last = informativeCount - 1;

   rightScale[last] = 1.0;
   singlepoint[last].UnPack();
   right[last].Copy(singlepoint[last]);
   singlepoint[last].Discard();

   bool   check_time = (mantra.bit_count >= 16) && (maxMinutes > 0);
   double seconds;
   time_t start_time, curr_time;

   if (check_time)
      {
      seconds = maxMinutes * 60.;
      // It would be good to check for problems in time call, but
      // errno is not actually part of ANSI specification
      // errno = 0;
      time(&start_time);
      // if (errno) check_time = false;
      }

   int rescalings = 0;

   for (int m = last - 1; m >= 0; m--)
      {
      ProgressReport("Right Conditional", last - m, informativeCount);

      // Retrieve recombination fraction
      double theta[2];

      theta[0] = keyFemaleTheta[m];
      theta[1] = keyMaleTheta[m];

      // Retrieve inheritance at m+1 given m + 1, m + 2, ...
      engine.Clear();
      engine.Copy(right[m+1]);
      right[m+1].Pack();

      if (useSparse && (informationBounds[m+1] + informationBounds[m] > 1.0))
         {
         // Try fast conditioning for informative locations?
         singlepoint[m].UnPack();
         right[m].Copy(singlepoint[m]);
         singlepoint[m].Discard();

         // Calculate conditional probability without intermediates
         engine.Condition(right[m], theta, rightScale[m+1]);
         }
      else
         {
         // Calculate inheritance at m given m+1, m+2, ...
         engine.MoveAlong(mantra, theta, rightScale[m+1]);

         // Calculate inheritance at m given *** m ***, m + 1, ...
         singlepoint[m].UnPack();
         engine.Multiply(singlepoint[m]);
         singlepoint[m].Discard();

         // Store it for future use
         right[m].Copy(engine);
         }

      // Check that the likelihod hasn't hit zero
      if (stats.GetMean(right[m]) == 0.0) return false;

      // If the tree is very unlikely, multiply by a constant to avoid underflow
      rightScale[m] = stats.mean < lowBound ? (rescalings+=m!=0, rescale) : 1.0;

      if (check_time)
         if (difftime(time(NULL), start_time) > seconds)
            throw OutOfTime();
      }

   if (informativeCount == 1) stats.GetMean(right[0]);
   likelihood += log(stats.mean) + rescalings * lnScale;

   right[0].Pack();
   return true;
   }

// Likelihood calculation assuming no recombination
//

void MerlinCore::SortInformation(int start_index, int end_index)
   {
   if (start_index >= end_index) return;

   double average = (informationScores[informationIndex[start_index]] +
                     informationScores[informationIndex[end_index]]) * 0.5;

   int a = start_index, b = end_index;

   while (true)
      {
      while ((informationScores[informationIndex[a]] >= average) && a < b)
         a++;

      while ((informationScores[informationIndex[b]] <= average) && a < b)
         b--;

      if (a == b) break;

      informationIndex.Swap(a, b);
      }

   SortInformation(start_index, a == end_index ? a - 1 : a);
   SortInformation(a == start_index ? a + 1 : a, end_index);
   }

bool MerlinCore::ZeroRecombination()
   {
   // Identify the most informative markers
   informationIndex.Dimension(informativeCount);
   informationIndex.SetSequence(0, 1);
   SortInformation(0, informativeCount - 1);

   // Combined tree includes all nodes that are non-zero in all trees
   Tree combined;

   // Allocate storage for final result
   right = new Tree[1];
   singlepoint[informationIndex[0]].UnPack();
   right[0].Copy(singlepoint[informationIndex[0]]);
   singlepoint[informationIndex[0]].Discard();

   // Keep track of how many likelihood rescalings were carried out
   int rescalings = 0;

   // Locate inheritance vectors that are compatible with all markers
   for (int i = 1; i < informativeCount; i++)
      {
      // Incorporate next most informative marker
      combined.Copy(right[0]);
      singlepoint[informationIndex[i]].UnPack();
      combined.Multiply(singlepoint[informationIndex[i]]);
      singlepoint[informationIndex[i]].Discard();
      combined.TrimNoMerge();
      right[0].Copy(combined);

      // Check that is possible given zero recombination
      if (stats.GetMean(right[0]) == 0.0)
         return false;

      // Avoid underflow by multiplying by large rescale constant
      if (stats.mean < lowBound)
         {
         rescalings++;
         right[0].Multiply(0, rescale);
         }
      }

   likelihood += log(stats.mean) + rescalings * lnScale;

   right[0].Pack();
   return true;
   }

// Calculation of right conditional likelihoods
//

void MerlinCore::WalkLeft
   (Multipoint & left, Multipoint & leftMarker, Multipoint & full, int marker)
   {
   // Load required information to memory
   singlepoint[marker+1].UnPack();
   right[marker+1].UnPack();

   // Calculate the new left conditioned probabilities
   if (marker != -1)
      {
      // First, retrieve the appropriate recombination fraction
      double theta[2];

      theta[0] = keyFemaleTheta[marker];
      theta[1] = keyMaleTheta[marker];

      // Load previous left conditioned likelihoods
      left.Copy(leftMarker);

      // Load the singlepoint likelihoods at the next location
      leftMarker.Copy(singlepoint[marker + 1]);

      if (!findErrors && useSparse &&
         (informationBounds[marker] + informationBounds[marker + 1] > 1.0))
         {
         // Build left conditioned probability up to marker + 1 (inclusive)
         left.Condition(leftMarker, theta, leftScale[marker]);
         leftScale[marker + 1] = stats.GetMean(leftMarker) < lowBound ? rescale : 1.0;

         // TODO -- We could choose to condition on left / right ?
         // Get right conditioned probability
         full.Copy(right[marker + 1]);
         left.Condition(full, theta, leftScale[marker]);
         }
      else
         {
          // Do the multipoint calculation
         left.MoveAlong(mantra, theta, leftScale[marker]);

         // Get the full left conditioned probability
         leftMarker.Multiply(left);
         leftScale[marker + 1] = stats.GetMean(leftMarker) < lowBound ? rescale : 1.0;

         // Get the full left and right conditioned probability
         full.Copy(right[marker + 1]);
         full.Multiply(left);
         }
       }
   else
      {
      // This is for the first marker
      leftMarker.Copy(singlepoint[0]);
      full.Copy(right[0]);
      leftScale[0] = 1.0;
      }

   // Check the marker for errors
   if (findErrors) ErrorCheck(left, full, marker + 1);

   // Free up memory
   singlepoint[marker+1].Discard();
   right[marker+1].Discard();
   }

// Scan chromosome and calculate likelihoods at each location for analyses
//

void MerlinCore::ScanChromosome()
   {
   // Total number of analysis locations
   int positionCount = analysisPositions.dim;

   if (positionCount == 0)
      return;

   // If no recombination ...
   if (zeroRecombination)
      {
      right[0].UnPack();
      AnalyseLocation(0, right[0]);
      right[0].Discard();

      return;
      }

   // Special case of two point analysis
   if (twopoint)
      {
      for (int pos = 0; pos < positionCount; pos++)
         {
         singlepoint[pos].UnPack();
         AnalyseLocation(pos, singlepoint[pos]);
         singlepoint[pos].Discard();
         }
      return;
      }

   // Left-conditioned probabilities up to left-marker
   Multipoint leftMarker;

   // Left and right conditionals up to current position
   Multipoint leftInheritance, inheritance;

   // Setup auxiliary variables for processing very sparse trees
   if (useSparse) leftInheritance.SetupConditioning(mantra);

   // Intermediate variables used for calculating recombination fractions
   double theta[2];

   // First we analyse any locations preceding the first informative marker
   int pos = 0;
   double firstMarkerPos = informativePositions[0];

   while (pos < positionCount && analysisPositions[pos] < firstMarkerPos)
      {
      ProgressReport("Scanning Chromosome", pos, positionCount);

      int marker = informativeMarkers[0];

      theta[0] = DistanceToRecombination(markerFemalePositions[marker] - femalePositions[pos]);
      theta[1] = DistanceToRecombination(markerMalePositions[marker] - malePositions[pos]);

      right[0].UnPack();
      inheritance.Copy(right[0]);
      inheritance.MoveAlong(mantra, theta);
      right[0].Discard();

      AnalyseLocation(pos++, inheritance);
      }

   // Next analyse positions, including marker locations and locations
   // between markers
   double lastMarkerPos = informativePositions.Last();

   // Anchors for starting position
   int leftAnchor  = -1;
   int rightAnchor =  0;

   while (pos < positionCount && analysisPositions[pos] <= lastMarkerPos)
      {
      ProgressReport("Scanning Chromosome", pos, positionCount);

      double position = analysisPositions[pos];

      // Move anchors so they flank the current position
      while (rightAnchor < informativeCount &&
             informativePositions[rightAnchor] <= position)
         {
         WalkLeft(leftInheritance, leftMarker, inheritance, leftAnchor++);
         rightAnchor++;
         }

      // If we are analysing at marker, the hard work is done
      if (informativePositions[leftAnchor] == position)
          {
          // nothing to do!
          }
      // For some types of analysis, we don't need between marker information
      else if (!betweenMarkers)
         {
         pos++;
         continue;
         }
      // Otherwise we are really between markers
      else
         {
         int marker = informativeMarkers[leftAnchor];

         theta[0] = DistanceToRecombination(femalePositions[pos] - markerFemalePositions[marker]);
         theta[1] = DistanceToRecombination(malePositions[pos] - markerMalePositions[marker]);

         leftInheritance.Copy(leftMarker);
         leftInheritance.MoveAlong(mantra, theta, leftScale[leftAnchor]);

         marker = informativeMarkers[rightAnchor];

         theta[0] = DistanceToRecombination(markerFemalePositions[marker] - femalePositions[pos]);
         theta[1] = DistanceToRecombination(markerMalePositions[marker] - malePositions[pos]);

         right[rightAnchor].UnPack();
         inheritance.Copy(right[rightAnchor]);
         inheritance.MoveAlong(mantra, theta, rightScale[rightAnchor]);
         inheritance.Multiply(leftInheritance);
         right[rightAnchor].Discard();
         }

      AnalyseLocation(pos++, inheritance);
      }

   if (pos < positionCount)
      while (rightAnchor++ < informativeCount)
         WalkLeft(leftInheritance, leftMarker, inheritance, leftAnchor++);

   // Finalize analyse any locations beyond the last informative marker
   while (pos < positionCount)
      {
      ProgressReport("Scanning Chromosome", pos, positionCount);

      int marker = informativeMarkers.Last();

      theta[0] = DistanceToRecombination(femalePositions[pos] - markerFemalePositions[marker]);
      theta[1] = DistanceToRecombination(malePositions[pos] - markerMalePositions[marker]);

      inheritance.Copy(leftMarker);
      inheritance.MoveAlong(mantra, theta);

      AnalyseLocation(pos++, inheritance);
      }
   }

// Output buffering functions
//

void MerlinCore::PrintHeader()
   {
   printHeader = false;

   printf("Family: %5s - Founders: %-2d - Descendants: %-2d - Bits: %-2d\n",
          (const char *) family->famid, mantra.f, mantra.n - mantra.f,
          mantra.bit_count);
   }

void MerlinCore::PrintMessage(const char * msg, ...)
   {
   if (printHeader) PrintHeader();

   va_list ap;

   va_start (ap, msg);
   vprintf(msg, ap);
   printf("\n");
   va_end(ap);
   }

void MerlinCore::ProgressReport(const char * msg, int done, int total)
   {
   if (quietOutput) return;

   long check = (0x10000 >> mantra.bit_count) - 1;
   if (mantra.bit_count > 16 || ((done & check ) == check))
      {
      if (printHeader)
         printf("Family %s (%d bits) -",
               (const char *) family->famid, mantra.bit_count);
      printf("  %s: %3d%%%20s\r", msg, done * 100 / total, "");
      fflush(stdout);
      cleanOutput = true;
      }
   }

void MerlinCore::CleanMessages()
   {
   if (cleanOutput)
      {
      printf("%70s\r", "");
      fflush(stdout);
      }

   if (!printHeader)
      printf("\n");
   }

// Enforce maximum number of alleles on markers
//

void MerlinCore::ValidateMarker(MarkerInfo * info)
   {
#ifdef __USE_LONG_INT
   if (info->freq.dim > 64)
       error("Please recode marker %s (and possibly others), which has %d alleles\n"
             "This version of merlin supports up to 63 alleles",
             (const char *) info->name, info->freq.dim - 1);
#else
   if (info->freq.dim > 32)
       error("Please recode marker %s (and possibly others), which has %d alleles\n"
             "This version of merlin handles 31 alleles or fewer per marker\n"
             "\n"
             "Alternatively, you could recompile Merlin with the __USE_LONG_INT\n"
             "option, which enables support for up to 63 alleles\n"
             "See the Merlin Makefile for details.\n",
             (const char *) info->name, info->freq.dim - 1);
#endif
   }

// Stub functions that should be replaced by derived classes

void MerlinCore::AnalyseLocation(int, Tree &) { }
void MerlinCore::ErrorCheck(Multipoint &, Multipoint &, int ) { }


 
 
