////////////////////////////////////////////////////////////////////// 
// merlin/MerlinFamily.cpp 
// (c) 2000-2003 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Sunday March 07, 2004
// 
 
#include "MerlinFamily.h"
#include "MerlinHaplotype.h"
#include "MerlinModel.h"
#include "InformationContent.h"
#include "MapFunction.h"
#include "MathConstant.h"
#include "Houdini.h"
#include "NPL-ASP.h"
#include "NPL-QTL.h"

// General analyses
bool FamilyAnalysis::estimateIBD = false;
bool FamilyAnalysis::estimateKinship = false;
bool FamilyAnalysis::estimateKinship15 = false;
bool FamilyAnalysis::estimateInformation = false;
bool FamilyAnalysis::estimateMatrices = false;
bool FamilyAnalysis::selectCases = false;
bool FamilyAnalysis::storeKinship = false;
bool FamilyAnalysis::calcLikelihood = false;

// Haplotyping
bool FamilyAnalysis::bestHaplotype = false;
int  FamilyAnalysis::sampledHaplotypes = 0;
bool FamilyAnalysis::allHaplotypes = false;

// Non-parametric linkage analysis
bool FamilyAnalysis::nplAll = false;
bool FamilyAnalysis::nplPairs = false;
bool FamilyAnalysis::nplQtl = false;
bool FamilyAnalysis::nplDeviates = false;

// Output options
bool FamilyAnalysis::simwalk2 = false;
bool FamilyAnalysis::perFamily = false;
bool FamilyAnalysis::writePDF = false;

FamilyAnalysis::FamilyAnalysis(Pedigree & pedigree) :
   MerlinCore(pedigree), kinship(mantra), kinship15(mantra),
   ibd(mantra), matrix(mantra), kac(pedigree, labels), hybrid(*this),
   pdf(analysisPositions)
   {
   nplCount = 0;
   npl = NULL;
   errorfile = NULL;
   taskList = NULL;
   }

FamilyAnalysis::~FamilyAnalysis()
   {
   if (npl != NULL) delete [] npl;

   // Delete generic analysis tasks
   AnalysisTask * current;

   while (taskList != NULL)
      {
      current = taskList;
      taskList = taskList->next;
      if (current != &kinship) delete current;
      }
   }

bool FamilyAnalysis::SelectFamily(Family * f, bool warnOnSkip)
   {
   if (MerlinCore::SelectFamily(f))
      // If analysis for this family is proceeding ...
      {
      if (estimateMatrices) matrix.SelectFamily(&ped, family);
      if (estimateIBD) ibd.SelectFamily(&ped, family);
      if (estimateKinship15) kinship15.SelectFamily(&ped, family);

      return true;
      }

   // Tell simwalk2 we have skipped this family
   if (simwalk2) hybrid.Skip();

   return false;
   }

void FamilyAnalysis::SetupGlobals()
   {
   MerlinCore::SetupGlobals();

   // How many NPL statistics to calculate
   nplCount = (nplAll + nplPairs) * ped.affectionCount +
              (nplQtl + nplDeviates) * ped.traitCount;

   if (npl != NULL) delete [] npl;

   nplInv.Dimension(nplCount);
   nplMean.Dimension(nplCount);
   npl = new Tree[nplCount];

   // Information Content Calculations requested?
   if (estimateInformation)
      NewTask(new InformationContent);

   // Parametric analyses requested?
   MerlinModel parametricModels;

   parametricModels.LoadModels();

   if (parametricModels.CountModels())
      NewTask(parametricModels.CreateTask());

   // Record how kinship information will be used
   if (estimateKinship || selectCases || storeKinship)
      {
      kinship.write = estimateKinship;
      kinship.store = storeKinship;
      kinship.selectCases = selectCases;

      NewTask(&kinship);
      }

   // Carry out between marker analyses?
   betweenMarkers = estimateIBD || estimateKinship15 || estimateMatrices ||
      nplCount != 0 || taskList != NULL;

   // Carry out analyses at each marker?
   scanChromosome = betweenMarkers || findErrors;

   // Perform right conditional calculations?
   multipoint     = scanChromosome || bestHaplotype && zeroRecombination ||
                    (sampledHaplotypes != 0) || allHaplotypes || calcLikelihood;

   // Merge gene flow outcomes with identical likelihoods, but possibly
   // implying different allelic states
   InheritanceTree::aggressiveMerging = !allHaplotypes;

   // Set up labels for NPL output
   kac.LabelPhenotypes(nplAll, nplPairs, nplQtl, nplDeviates);

   // Check that we have the correct type of data for simwalk2 analyses
   if (simwalk2 && ped.affectionCount != 1)
      error("Joint analysis with SimWalk2 requires pedigree with exactly one discrete trait\n"
            "The current pedigree includes %d discrete traits\n", ped.traitCount);

   // Information for generic tasks
   taskInfo.m = &mantra;
   taskInfo.drawPDF = writePDF;
   taskInfo.pdf = &pdf;
   taskInfo.perFamily = perFamily;
   }

int FamilyAnalysis::SetupMap(int chromosome)
   {
   int next_chromosome = MerlinCore::SetupMap(chromosome);

   // Setup chromosome labels for PDF output
   if (writePDF)
#ifdef __CHROMOSOME_X__
      pdf.SelectChromosomeX();
#else
      pdf.SelectChromosome(ped.GetMarkerInfo(markers[0])->chromosome);
#endif

   // Allocates array for storing NPL scores for each location in each family
   if (nplCount != 0) kac.Prepare(nplCount, ped.familyCount);

   // Initialize overall likelihood for this chromosome
   lkSum = 0.0;
   lkCount = 0;

   // Chromosome specific setup information for generic tasks
   taskInfo.positions = analysisPositions.dim;
   taskInfo.families = ped.familyCount;
   taskInfo.labels = &labels;

   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->Setup(taskInfo);

   return next_chromosome;
   }

void FamilyAnalysis::Analyse()
   {
   singlepoint = NULL;
   right = NULL;

   // Update task information
   taskInfo.famno = mantra.family->serial;
   taskInfo.famid = &(mantra.family->famid);

   // Setup for generic tasks
   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->SetupFamily(taskInfo);

   try
      {
      BasicTree::FreeSwap();

      ScoreSinglepoint();

      if (informativeCount == 0)
         SkipFamily();
      else
         {
         if (nplCount) ScoreNPL();
         if (multipoint)
            {
            if (ScoreConditionals())
               {
               if (scanChromosome) ScanChromosome();
               if (calcLikelihood && perFamily)
                  PrintMessage("  lnLikelihood = %.3f", likelihood);
               if (simwalk2) hybrid.Output();
               if (allHaplotypes) haplo.All(*this);
               for (int i = 0; i < sampledHaplotypes; i++) haplo.Sample(*this);
               if (bestHaplotype && zeroRecombination) haplo.MostLikely(*this);
               lkSum += likelihood;
               lkCount ++;
               }
            else
               {
               PrintMessage("  SKIPPED: Requires impossible recombination pattern");
               AbortAnalysis();
               }
            if (!twopoint) delete [] right;
            }
         if (bestHaplotype && !zeroRecombination) haplo.MostLikely(*this);
         }

      CleanMessages();
      delete [] singlepoint;
      }
   catch (const TreesTooBig & problem)
      {
      PrintMessage("  SKIPPED: At least %d megabytes needed", problem.memory_request);
      CleanMessages();

      FreeMemory();
      AbortAnalysis();
      }
   catch (const OutOfTime & problem)
      {
      PrintMessage("  SKIPPED: Analysis would require more than %d minutes\n", maxMinutes);
      CleanMessages();

      FreeMemory();
      AbortAnalysis();
      };
   };

void FamilyAnalysis::FreeMemory()
   {
   if (singlepoint != NULL) delete [] singlepoint;
   if (right != NULL) delete [] right;
   }

void FamilyAnalysis::AbortAnalysis()
   {
   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->SkipFamily();

   if (nplCount != 0) kac.SkipFamily(family->serial);
   if (simwalk2) nplInv[0] = nplInv[1] = 0.0;
   };

void FamilyAnalysis::SkipFamily()
   {
   // Print out overall gene-flow likelihood
   if (calcLikelihood && perFamily)
      PrintMessage("  lnLikelihood = %.3f", likelihood);

   // And update overall likelihood for the sample
   lkSum += likelihood;
   lkCount++;

   // If there is no information on this family
   // IBD and kinship coefficients / matrices are the same at all markers
   Tree dummy;
   dummy.MakeMinimalTree(1.0, mantra.bit_count);

   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->AnalyseUninformative(taskInfo, dummy);

   if (estimateIBD || estimateKinship15 || estimateMatrices)
      {
      double sum;

      double scale = pow(2.0, mantra.bit_count);

      if (estimateIBD) sum = ibd.Calculate(dummy, labels[0]) * scale;
      if (estimateMatrices ) sum = matrix.Calculate(dummy, labels[0]) * scale;
      if (estimateKinship15) sum = kinship15.Calculate(dummy, labels[0]) * scale;

      for (int i = 1; i < labels.Length(); i++)
         {
         if (estimateIBD) ibd.Output(labels[i], sum);
         if (estimateKinship15) kinship15.Output(labels[i], sum);
         if (estimateMatrices) matrix.Output(labels[i], sum);
         }
      }

   // Output dummy haplotype -- something to look at -- keeps people happy
   if (allHaplotypes || bestHaplotype) haplo.UninformativeFamily(*this);

   for (int i = 0; i < sampledHaplotypes; i++)
      haplo.UninformativeFamily(*this);

   // And tell Simwalk2 this family isn't very interesting (to us!)
   if (simwalk2) hybrid.Skip();
   }

void FamilyAnalysis::ScoreNPLBounds(int index)
   {
   stats.GetBounds(npl[index]);

   nplMean[index] = stats.mean;
   if (stats.var > TINY)
      {
      nplInv[index] = 1.0 / sqrt(stats.var);
      kac.min[index][serial] = (stats.min - stats.mean) * nplInv[index];
      kac.max[index][serial] = (stats.max - stats.mean) * nplInv[index];
      }
   else
      nplInv[index] = 0.0;

   npl[index].Pack();
   }

void FamilyAnalysis::ScoreNPL()
   {
   int index = 0;

   if (nplAll)
      {
      NPL_ALL engine;

      for (int a = 0; a < ped.affectionCount; a++)
         {
         ProgressReport("Scoring NPL-ALL statistic", a, ped.affectionCount);

         engine.ScoreNPL(mantra, a);
         npl[index].Copy(engine);
         ScoreNPLBounds(index++);
         }
      }

   if (nplPairs)
      {
      NPL_ASP_Tree * engine = simwalk2 ?
         (NPL_ASP_Tree *) new NPL_Pairs_For_Simwalk2 :
         (NPL_ASP_Tree *) new NPL_Pairs;

      for (int a = 0; a < ped.affectionCount; a++)
         {
         ProgressReport("Scoring NPL-Pairs statistic", a, ped.affectionCount);

         engine->ScoreNPL(mantra, a);
         npl[index].Copy(*engine);
         ScoreNPLBounds(index++);
         }
      }

   if (nplQtl)
      {
      NPL_QTL engine;

      for (int t = 0; t < ped.traitCount; t++)
         {
         ProgressReport("Scoring NPL-QTL statistic", t, ped.traitCount);

         engine.ScoreWithSampleMean(mantra, t);
         npl[index].Copy(engine);
         ScoreNPLBounds(index++);
         }
      }

   if (nplDeviates)
      {
      NPL_QTL engine;

      for (int t = 0; t < ped.traitCount; t++)
         {
         ProgressReport("Scoring NPL-Deviate statistic", t, ped.traitCount);

         engine.ScoreWithMeanZero(mantra, t);
         npl[index].Copy(engine);
         ScoreNPLBounds(index++);
         }
      }
   }

void FamilyAnalysis::AnalyseLocation(int pos, Tree & inheritance)
   {
   // The probability of the data summed all possible inheritance vectors
   double & lk = taskInfo.lk = -1.0;

   if (estimateKinship15) kinship15.Calculate(inheritance, labels[pos]);
   if (estimateMatrices) lk = matrix.Calculate(inheritance, labels[pos]);
   if (estimateIBD) lk = ibd.Calculate(inheritance, labels[pos]);

   // Execute generic analysis hooks
   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->Analyse(taskInfo, inheritance, pos);

   if (nplCount && lk == -1.0)
      lk = stats.GetMean(inheritance);

   for (int a = 0; a < nplCount; a++)
      // Only analyse informative families (eg. multiple related affecteds)
      if (nplInv[a] != 0.0)
         {
         npl[a].UnPack();
         kac.scores[a][serial][pos] =
            (inheritance.MeanProduct(npl[a])/lk - nplMean[a]) * nplInv[a];
         npl[a].Discard();
         }
   }

void FamilyAnalysis::ErrorCheck
     (Multipoint & left, Multipoint & full, int informativeMarker)
   {
   if (informativeCount <= 1) return;

   Multipoint      rightInheritance;
   FuzzyInheritanceTree alternative;

   double singlepoint_likelihood= stats.GetMean(singlepoint[informativeMarker]);
   double multipoint_likelihood = stats.GetMean(full);

   int marker = markers[informativeMarkers[informativeMarker]];

   if (++informativeMarker != informativeCount)
      {
      double theta[2];

      theta[0] = keyFemaleTheta[informativeMarker - 1];
      theta[1] = keyMaleTheta[informativeMarker - 1];

      right[informativeMarker].UnPack();
      rightInheritance.Copy(right[informativeMarker]);
      right[informativeMarker].Discard();
      rightInheritance.MoveAlong(mantra, theta, rightScale[informativeMarker]);
      if (informativeMarker != 1)
         rightInheritance.Multiply(left);
      }

   for (int error = family->first; error <= family->last; error++)
      if (ped[error].markers[marker].isKnown())
         {
         Alleles saved_genotype = ped[error].markers[marker];
         ped[error].markers[marker][0] = ped[error].markers[marker][1] = 0;

         alternative.FuzzyScoreVectors(mantra, marker);

         double alternative_likelihood = stats.GetMean(alternative);

         if (alternative_likelihood != singlepoint_likelihood)
            {
            double alternative_multipoint;

            if (informativeMarker == informativeCount)
               alternative_multipoint = alternative.MeanProduct(left);
            else
               alternative_multipoint = alternative.MeanProduct(rightInheritance);

            double score =
               (multipoint_likelihood / alternative_multipoint) /
               (singlepoint_likelihood / alternative_likelihood);

            if (score < 0.025)
               {
               PrintMessage("  %s genotype for individual %s is unlikely [%.3e]",
                           (const char *) ped.markerNames[marker],
                           (const char *) ped[error].pid, score);
               fprintf(errorfile, "%10s %10s %10s %#10.3g\n",
                       (const char *) family->famid,
                       (const char *) ped[error].pid,
                       (const char *) ped.markerNames[marker], score);
               }
            }

         ped[error].markers[marker] = saved_genotype;
         }
   }

void FamilyAnalysis::ShowLODs()
   {
   if (nplCount) kac.Analyse(pdf);
   if (storeKinship) vc.Analyse(*this);

   // Generic analysis hooks
   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->Report(taskInfo);
   }

void FamilyAnalysis::OpenErrorFile()
   {
   errorfile = fopen((const char *) (MerlinCore::filePrefix + ".err"), "wt");
   if (errorfile == NULL)
      error("Can't open error file [%s.err]\n", (const char *) MerlinCore::filePrefix);
   fprintf(errorfile, "%10s %10s %10s %10s\n",
      "FAMILY", "PERSON", "MARKER", "RATIO");
   }

void FamilyAnalysis::OpenPerFamilyFiles()
   {
   if (nplCount) kac.OpenPerFamilyFiles();
   if (storeKinship) vc.OpenPerFamilyFile();
   }

void FamilyAnalysis::SetupFiles()
   {
   if (findErrors) OpenErrorFile();
   if (perFamily) OpenPerFamilyFiles();
   if (estimateMatrices) matrix.OpenFile();
   if (estimateKinship15) kinship15.OpenFile();
   if (estimateIBD) ibd.OpenFile();
   if (bestHaplotype || (sampledHaplotypes != 0) || allHaplotypes) haplo.OpenFile();
   if (simwalk2) hybrid.OpenFiles();
   if (useSwap) BasicTree::SetupSwap();
   if (writePDF) pdf.OpenFile((const char *) (MerlinCore::filePrefix + ".pdf"));

   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->OpenFiles(taskInfo, MerlinCore::filePrefix);
   }

void FamilyAnalysis::CloseFiles()
   {
   if (findErrors)
      {
      printf("Unlikely genotypes listed in file [%s.err]\n", (const char *) MerlinCore::filePrefix);
      fclose(errorfile);
      }
   if (perFamily && nplCount) kac.ClosePerFamilyFiles();
   if (perFamily && storeKinship) vc.ClosePerFamilyFile();
   if (estimateMatrices) matrix.CloseFile();
   if (estimateKinship15) kinship15.CloseFile();
   if (estimateIBD) ibd.CloseFile();
   if (bestHaplotype || (sampledHaplotypes != 0) || allHaplotypes) haplo.CloseFile();
   if (simwalk2) hybrid.CloseFiles();
   if (useSwap) BasicTree::CloseSwap();
   if (writePDF) pdf.CloseFile();

   for (AnalysisTask * task = taskList; task != NULL; task = task->next)
      task->CloseFiles();
   }

void FamilyAnalysis::ShowLikelihood()
   {
   if (!calcLikelihood) return;

   printf("lnLikelihood for %d families (%d skipped) = %.3f\n\n",
          lkCount, ped.familyCount - lkCount, lkSum);
   }
 
void FamilyAnalysis::NewTask(AnalysisTask * task)
   {
   task->next = taskList;
   taskList = task;
   }
 
