////////////////////////////////////////////////////////////////////// 
// pedstats/pedstats.cpp 
// (c) 2000-2004 Goncalo Abecasis, 2002-2004 Jan Wigginton
// 
// This file is distributed as part of the PEDSTATS source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile PEDSTATS.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Tuesday February 10, 2004
// 
 
#include "pedstats.h"
#include "pedstats_pdf.h"

String pedName = "pedstats.ped";
String dataName = "pedstats.dat";
String ibdName = "pedstats.ibd";
String pdfName = "pedstats.pdf";

FilteredPedigree ped;
IBDTable ibd;
PedigreePairs all_pairs(ped);

bool hardy = false, hardyAll = false, hardyFounders = false;
bool showAll = false;
bool nocheck = false;
bool pairs = false;
bool byGender = false;
bool byFamily = false;
bool byAffection = false;
bool trimPedigree = false;
bool rewritePedigree = false;
bool writePDF = false;
bool writeFamilyPDF = false;
bool writeTraitPDF = false;
bool writeMarkerPDF = false;
bool writeAffPDF = false;

int trait_filter = 0;
int covariate_filter = 0;
int affection_filter = 0;

String  ageLabel = "";
double  shortGeneration = 13.0;
double  longGeneration = 70.0;
bool    birthYear = false;

int minGenos = 0;
int minPhenos = 0;
int minCovariates = 0;
String affectedFor = "";
bool filterOn = false;
String textFilterLabel, pdfFilterLabel;

int main(int argc, char * argv[])
   {
   #ifndef PEDVERSION
      printf("Pedigree Statistics\n(c) 1999-2004 Goncalo Abecasis, 2002-2004 Jan Wigginton\n");
   #else
      printf("Pedigree Statistics - " PEDVERSION "\n(c) 1999-2004 Goncalo Abecasis, 2002-2004 Jan Wigginton\n");
   #endif

   ReadParameters(argc, argv);

   ped.Prepare(dataName);
   ped.Load(pedName);
   if (trimPedigree) ped.Trim();

   ped.InitializeFilter();
   filterOn = (minGenos > 0 || minPhenos > 0 || minCovariates > 0 || affectedFor != "");
   if (filterOn)
      FilterPedigree(ped);

   ibd.Load(ped, ibdName);
   all_pairs.Build();

   if (ageLabel != "")
      CheckAges(all_pairs);

   FamilyStatistics();
   TraitStatistics(stdout, ped, all_pairs.sibs);
   CovarStatistics(stdout, ped, all_pairs.sibs);
   AffectionStatistics(stdout, ped);

   if (pairs) PairStatistics(stdout, ped, all_pairs);
   if (byFamily) ByFamilyStatistics();

   if (!nocheck) ped.InheritanceCheck();

   MarkerStatistics(stdout, ped);
   IBDStatistics();

   if (writePDF || writeTraitPDF || writeFamilyPDF || writeMarkerPDF || writeAffPDF)
      {
      PDF pdf;
      pdf.OpenFile((const char *) pdfName);
      pdf.page.SetSize(psLetterR);

      printf("\nWriting graphical output...\n");

      if (writePDF || writeFamilyPDF) GraphFamilyStatistics(pdf);
      if (writePDF || writeTraitPDF)
         {
         GraphTraitStatistics(pdf);
         GraphCovariateStatistics(pdf);
         }
      if (writePDF || writeAffPDF) GraphAffectionStatistics(pdf);
      if (writePDF || writeMarkerPDF)
         {
         GraphMarkerStatistics(pdf);
         if (hardy || hardyAll) HWStatistics(false, &pdf);
         if (hardy || hardyFounders) HWStatistics(true, &pdf);
         }
      printf("Graphical output written to %s\n", (const char *) pdfName);
      }
   else
      {
      if (hardy || hardyAll) HWStatistics(false);
      if (hardy || hardyFounders) HWStatistics(true);
      }

   return 0;
   };

int Generation(Person * p)
   {
   if (p->isFounder())
      return 0;

   int father = Generation(p->father);
   int mother = Generation(p->mother);

   return (mother > father ? mother : father) + 1;
   }

BEGIN_LONG_PARAMETERS(longParameters)
   LONG_PARAMETER_GROUP("Pedigree File")
      LONG_PARAMETER("ignoreMendelianErrors", &nocheck)
      LONG_PARAMETER("chromosomeX", &Pedigree::chromosomeX)
      LONG_PARAMETER("trim", &trimPedigree)
   LONG_PARAMETER_GROUP("Hardy-Weinberg")
      LONG_PARAMETER("hardyWeinberg", &hardy)
      LONG_PARAMETER("checkFounders", &hardyFounders)
      LONG_PARAMETER("checkAll", &hardyAll)
      LONG_PARAMETER("showAll", &showAll)
   LONG_PARAMETER_GROUP("Output")
      LONG_PARAMETER("pairStatistics", &pairs)
      LONG_PARAMETER("rewritePedigree", &rewritePedigree)
   LONG_PARAMETER_GROUP("Grouping")
      LONG_PARAMETER("bySex", &byGender)
      LONG_PARAMETER("byFamily", &byFamily)
   LONG_PARAMETER_GROUP("PDF Options")
      LONG_PARAMETER("pdf", &writePDF)
      LONG_PARAMETER("familyPDF", &writeFamilyPDF)
      LONG_PARAMETER("traitPDF", &writeTraitPDF)
      LONG_PARAMETER("affPDF", &writeAffPDF)
      LONG_PARAMETER("markerPDF", &writeMarkerPDF)
   LONG_PARAMETER_GROUP("Filter")
      LONG_INTPARAMETER("minGenos", &minGenos)
      LONG_INTPARAMETER("minPhenos", &minPhenos)
      LONG_INTPARAMETER("minCovariates", &minCovariates)
      LONG_STRINGPARAMETER("affectedFor", &affectedFor)
   END_LONG_PARAMETERS();


void ReadParameters(int argc, char * argv[])
   {
   ParameterList pl;

   pl.Add(new StringParameter('p', "Pedigree File", pedName));
   pl.Add(new StringParameter('d', "Data File", dataName));
   pl.Add(new StringParameter('i', "IBD File", ibdName));
   pl.Add(new StringParameter('a', "Adobe PDF File", pdfName));
   pl.Add(new StringParameter('x', "Missing Value Code", Pedigree::missing));
   pl.Add(new LongParameters("Additional Options", longParameters));

   pl.Read(argc, argv);
   pl.Status();
   }

void FamilyStatistics()
   {
   if (ped.familyCount == 0)
      {
      printf("WARNING - Pedigree is empty");
      return;
      }

   int families = 0, founders = 0, nonfounders = 0;
   int males = 0, females = 0;
   int maxgen = 0, mingen = 9999;
   int maxcount, mincount;

   int max_males = 0, max_females = 0;
   int min_males = 9999, min_females = 9999;

   double avecount = 0.0, avegen = 0.0;
   double ave_females = 0.0, ave_males = 0.0;

   IntArray generations(ped.familyCount);
   IntArray family_females(ped.familyCount);
   IntArray family_males(ped.familyCount);

   IntArray filtered_founders;
   filtered_founders.Dimension(ped.familyCount);
   filtered_founders.Zero();

   IntArray filtered_nonfounders;
   filtered_nonfounders.Dimension(ped.familyCount);
   filtered_nonfounders.Zero();

   IntArray filtered_fam;
   filtered_fam.Dimension(ped.familyCount);
   filtered_fam.Zero();

   for (int f = 0; f < ped.familyCount; f++)
      for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
         if (ped.filtered[i])
            ped[i].isFounder() ? filtered_founders[f]++ : filtered_nonfounders[f]++;

   maxcount = mincount = ped.families[0]->count - filtered_founders[0] - filtered_nonfounders[0];

   for (int f = 0; f < ped.familyCount; f++)
      {
      founders    += ped.families[f]->founders - filtered_founders[f];
      nonfounders += ped.families[f]->count - filtered_nonfounders[f] - ped.families[f]->founders;

      int curr_count = ped.families[f]->count - filtered_nonfounders[f] - filtered_founders[f];

      if (curr_count == 0)
        {
        filtered_fam[f] = 1;
        continue;
        }
      families ++;

      maxcount = max(curr_count, maxcount);
      mincount = min(curr_count, mincount);
      avecount += curr_count;

      generations[f] = 0;
      family_males[f] = 0;
      family_females[f] = 0;

      for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
         {
         generations[f] = max(Generation(ped.persons[i]) + 1, generations[f]);

         if (ped.filtered[i]) continue;

         if (ped[i].sex == SEX_FEMALE)
            {
            females++;
            family_females[f] ++;
            }
         else if (ped[i].sex == SEX_MALE)
           {
           family_males[f]++;
           males ++;
           }
         }

      max_males = max(family_males[f], max_males);
      min_males = min(family_males[f], min_males);
      max_females = max(family_females[f], max_females);
      min_females = min(family_females[f], min_females);

      maxgen = max(generations[f], maxgen);
      mingen = min(generations[f], mingen);
      avegen += generations[f];
      ave_females += family_females[f];
      ave_males += family_males[f];
      }

   double div = (double) max(families, 1);
   avegen /= div;
   avecount /= div;
   ave_females /= div;
   ave_males /= div;

   IntArray counts(maxcount + 1), gens(maxgen + 1);
   IntArray female_cts(max_females + 1), male_cts(max_males + 1);

   counts.Set(0);
   gens.Set(0);
   female_cts.Set(0);
   male_cts.Set(0);

   for (int f = 0; f < ped.familyCount; f++)
      {
      if (filtered_fam[f]) continue;

      counts[ped.families[f]->count - filtered_founders[f] - filtered_nonfounders[f]]++;
      gens[generations[f]]++;
      female_cts[family_females[f]]++;
      male_cts[family_males[f]]++;
      }

   IntArray idxcount(maxcount + 1), idxgen(maxgen + 1);
   IntArray idx_females(max_females + 1), idx_males(max_males + 1);

   MakeIndex(idxcount, counts);
   MakeIndex(idxgen, gens);
   MakeIndex(idx_females, female_cts);
   MakeIndex(idx_males, male_cts);

   double fdiv = max((double) families, 1.0) / (double) 100.0;

   printf("\n\n\nPEDIGREE STRUCTURE %s\n"
          "====================%s\n"
          "       Individuals: %d \n"
          "          Founders: %d founders, %d nonfounders\n"
          "            Gender: %d females, %d males\n",
          filterOn ?  "(AFTER FILTERING)" : "",
          filterOn ?  "=================" : "",
          founders + nonfounders, founders, nonfounders, females, males);

   printf("          Families: %d\n\n"
          "       Family Sizes\n"
          "           Average: %.2f (%d to %d)\n"
          "      Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n"
          "       Generations%s\n"
          "           Average: %.2f (%d to %d)\n"
          "      Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n",
          families,
          avecount, mincount, maxcount,
          idxcount[0], (double) counts[idxcount[0]] / fdiv,
          idxcount[1], (double) counts[idxcount[1]] / fdiv,
          idxcount[2], (double) counts[idxcount[2]] / fdiv,
          filterOn ? "*" : "",
          avegen, mingen, maxgen,
          idxgen[0], (double) gens[idxgen[0]] / fdiv,
          idxgen[1], (double) gens[idxgen[1]] / fdiv,
          idxgen[2], (double) gens[idxgen[2]] / fdiv);

   if (byGender)
      printf( "Females Per Family\n"
          "           Average: %.2f (%d to %d)\n"
          "      Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n"
          "   Males Per Family\n"
          "           Average: %.2f (%d to %d)\n"
          "      Distribution: %d (%.1f%%), %d (%.1f%%) and %d (%.1f%%)\n\n",
          ave_females, min_females, max_females,
          idx_females[0], (double) female_cts[idx_females[0]] / fdiv,
          idx_females[1], (double) female_cts[idx_females[1]] / fdiv,
          idx_females[2], (double) female_cts[idx_females[2]] / fdiv,
          ave_males, min_males, max_males,
          idx_males[0], (double) male_cts[idx_males[0]] / fdiv,
          idx_males[1], (double) male_cts[idx_males[1]] / fdiv,
          idx_males[2], (double) male_cts[idx_males[2]] / fdiv);

   if (filterOn)
      printf("\n* NOTE: Generation structure reported for filtered data may include\n"
              "  filtered individuals\n\n\n");

   printf("Checking family connectedness ... \n");
   int connected = true;
   for (int f = 0, groups; f < ped.familyCount; f++)
      if ( (groups = ped.families[f]->ConnectedGroups()) != 1 )
        printf("  Family %s (%d individuals) consists of %d disjoint groups\n",
          (const char *) ped.families[f]->famid, ped.families[f]->count, groups),
        connected = false;

   printf(connected ? "  All individuals in each family are connected.\n\n" :
                      "  Use the --rewritePedigree option to break families"
                      " into connected groups\n\n");

   if (rewritePedigree)
      {
      printf("Rewriting data file as [pedstats.dat]... ");
      ped.WriteDataFile("pedstats.dat");

      printf("Done!\n"
             "Rewriting pedigree file as [pedstats.ped]... ");
      FILE * pedfile = fopen("pedstats.ped", "wt");

      IntArray groups;
      for (int f = 0; f < ped.familyCount; f++)
         {
         ped.families[f]->ConnectedGroups(&groups);

         for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
            ped.WritePerson(pedfile, i,
                            ped[i].famid + "_" + groups[ped[i].traverse]);
         }

      printf("Done!\n\n");
      fclose(pedfile);
      }
   }

void TraitStatistics(FILE * file, FilteredPedigree & ped, SimplePairList & pairs, const char * added_label)
   {
   if (Person::traitCount == 0) return;

   String extra = added_label != NULL ? added_label : "";
   if (filterOn)
      extra += "*";

   String fill('=', extra.Length() + 1);

   fprintf(file, "\n\n\nQUANTITATIVE TRAIT STATISTICS %s\n"
           "=============================%s\n",
           (const char *) extra, (const char *) fill);

   GetTraitStatistics(file, ped, pairs, SEX_UNKNOWN, "ALL DATA:", false);

   if (byGender)
      {
      GetTraitStatistics(file, ped, pairs, SEX_FEMALE, "FEMALES:", false);
      GetTraitStatistics(file, ped, pairs, SEX_MALE, "MALES:", false);
      }

   if (filterOn)
      fprintf(file, "\n* NOTE:  Sample restricted to %s\n", (const char *) textFilterLabel);
   }

void CovarStatistics(FILE * file, FilteredPedigree & ped, SimplePairList & pairs, const char * added_label)
   {
   if (Person::covariateCount == 0) return;

   String extra = added_label != NULL ? added_label : "";
   if (filterOn)
      extra += "*";
   String fill('=', extra.Length() + 1);

   fprintf(file, "\n\n\nCOVARIATE STATISTICS%s\n"
           "====================%s\n",
           (const char *) extra, (const char *) fill);

   GetTraitStatistics(file, ped, pairs, SEX_UNKNOWN, "ALL DATA:", true);

   if (byGender)
      {
      GetTraitStatistics(file, ped, pairs, SEX_FEMALE, "FEMALES:", true);
      GetTraitStatistics(file, ped, pairs, SEX_MALE, "MALES:", true);
      }

   if (filterOn)
      fprintf(file, "\n* NOTE:  Sample restricted to %s\n", (const char *) textFilterLabel);
   }

void GetTraitStatistics(FILE * file, FilteredPedigree & ped, SimplePairList & sib_pairs,
int sex_filter, const char * label, bool is_covariate)
   {
   if (byGender)
      fprintf(file, "\n%-15s\n", label);

   fprintf(file, "\n%11s %19s %8s %8s %8s %8s %8s\n",
           "", "[All Phenotypes]", "Min", "Max", "Mean", "Var", "SibCorr");

   int count = 0, possible = 0, fcount = 0, fpossible = 0;
   int filter_limit = is_covariate ? Person::covariateCount : Person::traitCount;

   StringArray founder_stats;
   founder_stats.Dimension(filter_limit);

   for (int t = 0; t < filter_limit; t++)
      {
      trait_filter = t;
      int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;

      double mean = 0.0, var = 0.0;
      double min_value = 1e300, max_value = -1e300;

      double mean_f = 0.0, var_f = 0.0;
      double min_fvalue = 1e300, max_fvalue = -1e300;

      bool   typed;
      double val;
      for (int i = 0; i < ped.count; i++)
         {
         if (ped.filtered[i]) continue;

         if (sex_filter != SEX_UNKNOWN  && ped[i].sex != sex_filter) continue;

         tpossible ++;
         if (ped[i].isFounder()) tfpossible ++;

         typed = is_covariate ? ped[i].isControlled(t) : ped[i].isPhenotyped(t);
         if (!typed ) continue;

         val = is_covariate ? ped[i].covariates[t] : ped[i].traits[t];

         min_value = min(min_value, val);
         max_value = max(max_value, val);
         mean    += val;
         var     += val * val;

         if (ped[i].isFounder())
            {
            tfcount ++;
            min_fvalue = min(min_fvalue, val);
            max_fvalue = max(max_fvalue, val);
            mean_f += val;
            var_f  += val * val;
            }
         tcount ++;
         }

      possible  += tpossible;
      fpossible += tfpossible;
      count     += tcount;
      fcount    += tfcount;

      var = (tcount <= 1) ? 0.0 : (double) (tcount * var - mean * mean) /(double)((tcount) * (tcount - 1.0));
      mean /= max(tcount, 1);

      var_f = (tfcount <= 1) ? 0.0 : (double) (tfcount * var_f - mean_f * mean_f) /(double)((tfcount) * (tfcount - 1.0));
      mean_f /= max(tfcount, 1);

      int num_pairs = 0;
      double r_sib;
      if (sex_filter != SEX_UNKNOWN)
         {
         SimplePairList filtered_sibs(ped);
         sex_filter == SEX_MALE ?
                       sib_pairs.FilterOnBothMale(filtered_sibs) :
                       sib_pairs.FilterOnBothFemale(filtered_sibs);

         r_sib = filtered_sibs.Correlation(t, is_covariate, true, num_pairs);
         }
      else
         r_sib = sib_pairs.Correlation(t, is_covariate, true, num_pairs);

      String trait_name = is_covariate ? Person::covariateNames[t] : Person ::traitNames[t];

      if (num_pairs > 1)
         fprintf(file, "%15.15s %8d %5.1f%% %8.*f %8.*f %8.*f %8.*f %8.3f\n",
                 (const char * ) trait_name, tcount,
                 (double) tcount / (double)  max(tpossible, 1) * 100,
                 fabs(min_value) < 1e4 ? 3 : 0, min_value,
                 fabs(max_value) < 1e4 ? 3 : 0, max_value,
                 fabs(mean) < 1e4      ? 3 : 0, mean,
                 var < 1e4  ? 3 : 0, var, r_sib);
      else if (tcount > 0)
         fprintf(file, "%15.15s %8d %5.1f%% %8.*f %8.*f %8.*f %8.*f %8s\n",
                 (const char * ) trait_name, tcount,
                 (double) tcount / (double)  max(tpossible, 1) * 100,
                  fabs(min_value) < 1e4 ? 3 : 0, min_value,
                  fabs(max_value) < 1e4 ? 3 : 0, max_value,
                  fabs(mean) < 1e4      ? 3 : 0, mean,
                  var < 1e4  ? 3 : 0, var, "-");
      else
         fprintf(file, "%15.15s %8d %5.1f%% %8s %8s %8s %8s %8s\n",
                 (const char * ) trait_name, tcount,
                 (double) tcount / (double)  max(tpossible, 1) * 100,
                 "-", "-", "-", "-", "-");

      if (tfcount > 0)
         founder_stats[t].printf("%15.15s %8d %5.1f%% %8.*f %8.*f %8.*f %8.*f %8s",
            (const char * ) trait_name, tfcount,
            (double) tfcount / (double)  max(tfpossible, 1) * 100,
            fabs(min_fvalue) < 1e4 ? 3 : 0, min_fvalue,
            fabs(max_fvalue) < 1e4 ? 3 : 0, max_fvalue,
            fabs(mean_f) < 1e4     ? 3 : 0, mean_f,
            var_f < 1e4 ? 3 : 0, var_f, "-");
      else
         founder_stats[t].printf("%15.15s %8d %5.1f%% %8s %8s %8s %8s %8s",
            (const char * ) trait_name, tfcount,
            (double) tfcount / (double)  max(tfpossible, 1) * 100,
            "-", "-", "-", "-", "-");
      }

   fprintf(file, "%15.15s %8d %5.1f%%\n\n", "Total",
           count,  (double) count / (double) max(possible, 1) * 100);

   fprintf(file, "\n%10s %20s %8s %8s %8s %8s %8s\n",
          "", "[Founders Only]", "Min", "Max", "Mean", "Var", "SibCorr");

   for (int i = 0; i < filter_limit; i++)
      fprintf(file, "%60s\n", (const char *) founder_stats[i]);

   fprintf(file, "%15.15s %8d %5.1f%%\n\n", "Total",
           fcount,  (double) fcount  / (double) max(fpossible, 1) * 100);
   }

void AffectionStatistics(FILE * file, FilteredPedigree & ped, const char * added_label)
   {
   if (Person::affectionCount == 0) return;

   String extra = added_label != NULL ? added_label : "";
   if (filterOn)
      extra += "*";
   String fill('=', extra.Length() + 1);

   fprintf(file, "\n\n\nAFFECTION STATISTICS%s\n" "====================%s\n",
          (const char *) extra, (const char *) fill);

   GetAffectionStatistics(file, ped, SEX_UNKNOWN, "ALL DATA:");

   if (byGender)
      {
      GetAffectionStatistics(file, ped, SEX_FEMALE, "FEMALES:");
      GetAffectionStatistics(file, ped, SEX_MALE, "MALES:");
      }

   if (filterOn)
      fprintf(file, "\n* NOTE:  Sample restricted to %s\n", (const char *) textFilterLabel);
   }

void GetAffectionStatistics(FILE * file, FilteredPedigree & ped, int sex_filter, const char * label)
   {
   if (byGender) fprintf(file, "\n%-15s", label);
   fprintf(file, "\n%15s %15s %15s %10s\n", "", "[Diagnostics]", "[Founders]", "Prevalence");

   int count = 0, possible = 0, fcount = 0, fpossible = 0;

   for (int a = 0; a < Person::affectionCount; a++)
      {
      int    tcount = 0, tpossible = 0, tfcount = 0, tfpossible = 0;
      int    affecteds = 0;

      for (int i = 0; i < ped.count; i++)
         {
         if (ped.filtered[i]) continue;

         if (sex_filter != SEX_UNKNOWN && ped[i].sex != sex_filter) continue;
         tpossible  ++;
         tfpossible += ped[i].isFounder();

         if (!ped[i].isDiagnosed(a)) continue;

         affecteds += ped[i].affections[a] == 2;
         tfcount   += ped[i].isFounder();
         tcount ++;
         }

      possible  += tpossible;
      fpossible += tfpossible;
      count     += tcount;
      fcount    += tfcount;

      fprintf(file, "%15.15s %8d %5.1f%% %8d %5.1f%% %9.1f%%\n",
              (const char *) Person::affectionNames[a],
              tcount, (double) tcount / (double)  max(tpossible, 1) * 100,
              tfcount, (double) tfcount / (double) max(tfpossible, 1) * 100,
              (double) affecteds / (double) max(tcount, 1) * 100 );
      }

   fprintf(file, "%15s %8d %5.1f%% %8d %5.1f%%\n\n", "Total",
           count,  (double) count  / (double) max(possible, 1) * 100,
           fcount, (double) fcount / (double) max(fpossible, 1) * 100 );
   }

void MarkerStatistics(FILE * file, FilteredPedigree & ped, const char * added_label)
   {
   if (Person::markerCount == 0) return;

   String extra = added_label != NULL ? added_label : "";
   if (filterOn)
      extra += "*";
   String fill('=', extra.Length() + 1);

   fprintf(file, "\n\n\nMARKER GENOTYPE STATISTICS%s\n"
          "==========================%s\n",
          (const char *) extra, (const char *) fill);

   GetMarkerStatistics(file, ped, SEX_UNKNOWN, "ALL DATA:");

   if (byGender)
      {
      GetMarkerStatistics(file, ped, SEX_FEMALE, "FEMALES:");
      GetMarkerStatistics(file, ped, SEX_MALE, "MALES:");
      }

   if (filterOn)
      fprintf(file, "\n* NOTE:  Sample restricted to %s\n", (const char *) textFilterLabel);
   }

void GetMarkerStatistics(FILE * file, FilteredPedigree & ped, int sex_filter, const char * label)
   {
   if (byGender) fprintf(file, "%-15s", label);

   fprintf(file, "\n%15s %15s %15s %10s\n",
          " ", "[Genotypes]", "[Founders]", "Hetero");

   int genotypes = 0, possible = 0, fgenotypes = 0, fpossible = 0,
       hetero = 0, maxHetero = 0;

   for (int m = 0; m < Person::markerCount; m++)
      {
      int mGenotypes = 0, mPossible = 0, mfGenotypes = 0, mfPossible = 0,
          mHetero = 0, mMaxHetero = 0;
      int m0, m1;
      bool is_founder;

      for (int i = 0; i < ped.count; i++)
         {
         if (ped.filtered[i]) continue;
         if (sex_filter != SEX_UNKNOWN && ped[i].sex != sex_filter) continue;

         m0 = ped[i].markers[m].one;
         m1 = ped[i].markers[m].two;
         is_founder = ped[i].isFounder();

         mPossible ++;
         mfPossible += is_founder;

         if (m0 == 0 || m1 == 0) continue;

         mGenotypes ++;
         mfGenotypes += is_founder;

         if (ped.chromosomeX && ped[i].sex == SEX_MALE) continue;
         if (m0 != m1)
            mHetero ++;
         mMaxHetero ++;
         }

      possible   += mPossible;
      fpossible  += mfPossible;
      fgenotypes += mfGenotypes;
      genotypes  += mGenotypes;
      hetero     += mHetero;
      maxHetero  += mMaxHetero;

      fprintf(file, "%15.15s %8d %5.1f%% %8d %5.1f%% %9.1f%%\n",
              (const char *) Person::markerNames[m],
              mGenotypes, (double) mGenotypes / (double) max(mPossible, 1) * 100,
              mfGenotypes, (double)mfGenotypes / (double)max(mfPossible, 1) * 100,
              (double) mHetero / (double) max(mMaxHetero, 1) * 100);
      }

   fprintf(file, "%15s %8d %5.1f%% %8d %5.1f%% %9.1f%%\n%s\n", "Total",
           genotypes, (double) genotypes / (double) max(possible, 1) * 100,
           fgenotypes, (double) fgenotypes / (double) max(fpossible, 1) * 100,
           (double) hetero / (double) max(maxHetero, 1) * 100,
           ped.chromosomeX ? "\nNOTE: Heterozygosity calculation excludes males\n" : "");
   }

void IBDStatistics()
   {
   if (Person::markerCount == 0) return;
   if (ibd.isEmpty()) return;

   printf("\n\n\nIBD FAMILY INFORMATION (FROM FILE)\n=================================\n\n");

   printf("%17s%10s %10s  %16s  %16s\n",
          " ", "[Singletons]", "[Trios]", "[2 generations]", "[3+ generations]");

   for (int m = 0; m < Person::markerCount; m++)
      {
      int ibdTwoGen = 0,  ibdPossTwoGen = 0;
      int ibdTwoPlusGen = 0, ibdPossTwoPlusGen = 0;
      int ibdTrios = 0, ibdPossTrios = 0;
      int ibdSingletons = 0, ibdPossSingletons = 0;

      for (int f = 0; f < ped.familyCount; f++)
         {
         bool have_family = ibd.HaveFamily(m, ped.families[f]) && !ibd.isEmpty();

         if (ped.families[f]->count == 3 && ped.families[f]->founders == 2)
            {
            if (have_family) ibdTrios ++;
            ibdPossTrios ++;
            }
         else if (ped.families[f]->count == ped.families[f]->founders)
            {
            if (have_family) ibdSingletons ++;
            ibdPossSingletons ++;
            }

         int generations = 0;
         for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
            {
            generations = max(Generation(ped.persons[i]) + 1, generations);
            }

         if (generations == 2)
            {
            if (have_family) ibdTwoGen ++;
            ibdPossTwoGen ++;
            }
         else if (generations > 2)
            {
            if (have_family) ibdTwoPlusGen ++;
            ibdPossTwoPlusGen ++;
            }
         }
      printf("%15s %6d/%-6d %6d/%-6d %7d/%-8d %7d/%-8d\n",
             (const char *) Person::markerNames[m],
             ibdSingletons, ibdPossSingletons, ibdTrios, ibdPossTrios,
             ibdTwoGen, ibdPossTwoGen, ibdTwoPlusGen, ibdPossTwoPlusGen);
     }

   printf("\n\nNOTE: All counts are number of families per category that have some\n"
          "      marker IBD information available.\n\n\n");

   if (filterOn)
      printf("\n* NOTE:  Generation structure reported includes filtered individuals\n\n");
   }

void MakeIndex(IntArray & index, IntArray & counts)
   {
   index.Dimension(counts.Length());
   index.SetSequence(0, 1);

   // Insertion sort should be OK, for N < 20
   for (int i, j = 1; j < counts.Length(); j++)
      {
      int tmp = counts[j];    // index[j] = j
      for (i = j - 1; (i >= 0) && (counts[index[i]] < tmp); i--)
         index[i + 1] = index[i];
      index[i + 1] = j;
      }

   // Make sure we have at least three items
   for (int i = counts.Length() + 1; i < 4; i++)
      {
      counts.Dimension(i);
      index.Dimension(i);

      counts[i - 1] = 0;
      index[i - 1] = i - 1;
      }
   }

void HWStatistics(bool foundersOnly, PDF * pdf)
   {
   if (ped.markerCount == 0) return;

   bool fail = false;
   bool note = false;

   if (ped.chromosomeX)
      printf("\n\n\nHARDY-WEINBERG CHECK AMONG %sFEMALES%s\n"
             "============================%s========\n",
             foundersOnly ? "FOUNDER " : "",
             filterOn ? "*" : "",
             foundersOnly ? "========" : "");

   else
      printf("\n\n\nHARDY-WEINBERG CHECK%s%s\n"
             "====================%s\n",
             foundersOnly ? " AMONG FOUNDERS" : "",
             filterOn ? "*" : "",
             foundersOnly ? "===============" : "");

   Matrix   observed;
   Vector   freqs;
   IntArray rareAlleles;
   StringArray sparseNames;

   int tested_markers = 0;

   for (int m = 0; m < Person::markerCount; m++)
      {
      int alleleCount = ped.CountAlleles(m);

      if (alleleCount <= 1) continue;

      observed.Dimension(alleleCount, alleleCount);
      observed.Zero();

      freqs.Dimension(alleleCount);
      freqs.Zero();

      rareAlleles.Dimension(alleleCount);
      rareAlleles.Zero();

      int total = 0;

      // First calculate allele frequencies
      for (int i = 0; i < ped.count; i++)
         {
         if (ped.filtered[i]) continue;
         if (ped[i].isGenotyped(m) &&
            (!ped.chromosomeX || ped[i].sex != SEX_MALE) &&
            (!foundersOnly || ped[i].isFounder()))
            {
            int one = ped[i].markers[m].one - 1;
            int two = ped[i].markers[m].two - 1;

            freqs[one] ++;
            freqs[two] ++;
            total ++;
            }
         }

      if (total <= 3.0)
         {
         sparseNames.Push(Person::markerNames[m]);
         continue;
         }

      // Work out allele frequencies
      freqs.Multiply(1.0 / (2.0 * total));

      // Find the minimum useful allele frequency
      double minFreq = sqrt(3.0 / total);

      // List all rare alleles and find their total frequency
      double rareSum = 0.0;
      String rareNames = "";

      // Also find the rarest allele with frequency >= minFreq
      int mini = -1, commonCount = 0;

      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] == 0.0)
            continue;
         else if (freqs[i] < minFreq)
            {
            rareSum += freqs[i];
            rareAlleles[i] = true;
            rareNames += " ";
            rareNames += (i + 1);
            }
         else
            {
            if (mini == -1 || freqs[i] < freqs[mini])
               mini = i;
            commonCount ++;
            }

      // If all rare alleles combined are still too rare...
      if (rareSum > 0.0 && rareSum < minFreq && commonCount > 1)
         {
         rareSum += freqs[mini];
         rareAlleles[mini] = true;
         rareNames += " ";
         rareNames += (mini + 1);
         }

      // Glob all rare alleles into 1
      int glob = alleleCount - 1;

      if (rareSum > 0.0)
         while (rareAlleles[glob] == false)
            glob--;

      // Recount allele and genotype frequencies globbing rare alleles as needed
      freqs.Zero();

      for (int i = 0; i < ped.count; i++)
         {
         if (ped.filtered[i]) continue;

         if (ped[i].isGenotyped(m) &&
            (!ped.chromosomeX || ped[i].sex != SEX_MALE) &&
            (!foundersOnly || ped[i].isFounder()))
            {
            int one = ped[i].markers[m].two - 1;
            int two = ped[i].markers[m].one - 1;

            if (rareAlleles[one]) one = glob;
            if (rareAlleles[two]) two = glob;

            if (one > two)
               observed[one][two] ++;
            else
               observed[two][one] ++;

            freqs[one] ++;
            freqs[two] ++;
            }
         }
      freqs.Multiply(1.0 / (2.0 * total));

      double chisq = 0.0;
      bool   low_counts = false;

      int realAlleles = 0;
      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] != 0.0)
             realAlleles ++;

      StringArray alleleNames;
      alleleNames.Dimension(realAlleles);

      Matrix real_expected, real_observed;
      real_expected.Dimension(realAlleles, realAlleles);
      real_observed.Dimension(realAlleles, realAlleles);
      real_expected.Zero();
      real_observed.Zero();

      int j = 0;
      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] != 0.0)
            {
            if (i == glob && rareSum > 0.0)
               alleleNames[j] = "* P";
            else
               alleleNames[j].printf("%d", i + 1);
            j ++;
            }

      double expected;
      int l = 0, k = 0;
      for (int i = 0; i < alleleCount; i++)
         if (freqs[i] > 0)
            {
            l = 0;
            for (int j = 0; j <= i; j++)
               if (freqs[j] > 0)
                  {
                  expected = freqs[i] * freqs[j] * total * (i == j ? 1 : 2);
                  chisq += square(expected - observed[i][j]) / expected;
                  low_counts |= expected < 3.0;
                  real_expected[k][l] = expected;
                  real_observed[k][l] = observed[i][j];
                  l ++;
                  }
            k ++;
            }

      note |= low_counts;

      tested_markers ++;
      double pvalue = realAlleles == 1.0 ? 1.0 :
                      chidist(chisq, realAlleles * (realAlleles - 1) * 0.5);

      String founder_label;

      if (ped.chromosomeX)
         founder_label.printf("( Among %s )", foundersOnly ? "Female Founders" : "Females");
      else
         founder_label.printf("%s", foundersOnly ? "Among Founders" : "");
      
      
      if (pvalue > 0.05 && !showAll) continue;

      if (pdf != NULL)
         GraphHardyWeinberg(*pdf, real_observed, real_expected,
          (const char *) Person::markerNames[m], alleleNames, chisq, pvalue,
          (const char *) founder_label, (const char *) rareNames);

      if (!fail)
         printf("%15s %7s  %7s %7s %7s %7s\n",
                "", "N", "Alleles", "Largest", "ChiSq", "P-value");
      fail = true;

      printf("%15.15s %7d  %7d %7d %7.2f %7.4f %s",
             (const char *) Person::markerNames[m],
             total, realAlleles, alleleCount, chisq, pvalue,
             low_counts ? "see note " : "");

      if (rareAlleles.Sum() > 1)
         printf("(%d alleles grouped)\n", rareAlleles.Sum());
      else
         printf("\n");
      }

   if (note && fail)
      printf("\nNOTE: Some genotypes had very low expecteds counts, making this HW test \nunreliable.\n");

   if (tested_markers == 0)
      printf("No markers tested.\n\n");
   else
      {
      if (tested_markers == 1)
         printf(!fail ? "All tested markers okay.\n\n" : "\n");
      else
         printf(!fail ? "All %d tested markers okay.\n\n" : "\n", tested_markers);
      }

   if (filterOn)
      printf("\n* NOTE:  Sample restricted to %s\n\n", (const char *) textFilterLabel);

   int sparse_ct = sparseNames.Length();
   if (sparse_ct > 0)
      {
      printf("The following markers were not tested for Hardy-Weinberg "
              "due to a low \nnumber of genotyped individuals:\n");

      for (int i = 0, line = 80; i < sparse_ct; i++)
         {
         if (line + sparseNames[i].Length() + 1 > 79)
            printf("\n   "), line = 3;

         printf("%s ", (const char *) sparseNames[i]);
         line += sparseNames[i].Length() + 1;
         }

      printf("\n\n");
      }
   }

void PairStatistics(FILE * file, FilteredPedigree & ped, PedigreePairs & pairs, const char * added_label,
bool traits_only)
   {
   String extra = added_label != NULL ? added_label : "";
   if (filterOn)
      extra += "*";
   String fill('=', extra.Length() + 1);

   fprintf(file, "\n\n\nPAIR STATISTICS%s\n" "================%s\n",
          (const char *) extra, (const char *) fill);

   GetPairStatistics(file, pairs, "ALL DATA:", traits_only);

   PedigreePairs female_prs(ped), male_prs(ped), opp_prs(ped);
   pairs.SplitOnGender(female_prs, male_prs, opp_prs);

   if (byGender)
      {
      GetPairStatistics(file, female_prs, "SAME SEX PAIRS (FEMALE):", traits_only);
      GetPairStatistics(file, male_prs, "SAME SEX PAIRS (MALE):", traits_only);
      GetPairStatistics(file, opp_prs, "OPPOSITE SEX PAIRS:", traits_only);
      }

   if (filterOn)
      fprintf(file, "\n* NOTE:  Sample restricted to %s\n", (const char *) textFilterLabel);
   }

void GetPairStatistics(FILE * file, PedigreePairs & pairs, const char * label, bool traits_only)
   {
   if (byGender)
      fprintf(file, "\n\n%-15s %s\n", label, pairs.totalCt > 0 ? "" : "NONE" );
   else if (pairs.totalCt <= 0)
      fprintf(file, "\n\n%s\n", "NO RELATIVE PAIRS");

   if (pairs.totalCt <= 0) 
      return;

   if (!traits_only)
      pairs.PrintCounts(file);

   PairTraitStatistics(file, pairs, false);
   PairTraitStatistics(file, pairs, true);

   if (!traits_only)
     PairAffectionStatistics(file, pairs);
   }

void PairAffectionStatistics(FILE * file, PedigreePairs & pairs)
   {
   if (Person::affectionCount == 0) return;

   fprintf(file, "\n\nPair Counts by Affection Status:\n");

   fprintf(file, "%15s %8s %8s %8s %12s %12s %9s", " ",
          "Sib", "HalfSib","Cousin", "ParentChild", "Grandparent", "Avuncular");

   for (affection_filter = 0; affection_filter < Person::affectionCount; affection_filter++)
      {
      IntArray unaff, disc, aff;

      pairs.CountAffectionTypes(affection_filter, unaff, disc, aff);

      fprintf(file, "\n%15.15s\n", (const char *) ped.affectionNames[affection_filter]);

      fprintf(file, "%2s[%11s] %8d %8d %8d %12d %12d %9d\n", " ", "Unaffected",
              unaff[0], unaff[1], unaff[2], unaff[3], unaff[4], unaff[5]);

      fprintf(file, "%2s[%11s] %8d %8d %8d %12d %12d %9d\n", " ", "Discordant",
              disc[0], disc[1], disc[2], disc[3], disc[4], disc[5]);

      fprintf(file, "%2s[%11s] %8d %8d %8d %12d %12d %9d\n", " ", "Affected",
              aff[0], aff[1], aff[2], aff[3], aff[4], aff[5]);
      }

   if (pairs.sibs.gender == "Opposite")
      {
      bool header_printed = false;

      for (affection_filter = 0; affection_filter < ped.affectionCount; affection_filter++)
         {
         PedigreePairs mf_pairs(ped), fm_pairs(ped);

         if ( pairs.avuncularCt + pairs.grandParentCt + pairs.parentCt <= 0 ) continue;

         pairs.SplitOnGenderOrder(mf_pairs, fm_pairs);

         IntArray mf_unaff, mf_disc, mf_aff;
         IntArray fm_unaff, fm_disc, fm_aff;

         mf_pairs.CountAffectionTypes(affection_filter, mf_unaff, mf_disc, mf_aff, false);
         fm_pairs.CountAffectionTypes(affection_filter, fm_unaff, fm_disc, fm_aff, false);

         int total_diagnosed = 0;
         for (int i = 3; i < 6; i++)
             total_diagnosed += mf_unaff[i] + mf_disc[i] + mf_aff[i]
                             +  fm_unaff[i] + fm_disc[i] + fm_aff[i];
         if (total_diagnosed == 0) continue;

         if (header_printed == false) PrintDetailHeader(file);
         header_printed = true;

         fprintf(file, "%15.15s\n", (const char *) Person::affectionNames[affection_filter]);

         fprintf(file, "%2s[%11s] %8d %8d %8d %8d %8d %8d\n", " ", "Unaffected",
           mf_unaff[3], fm_unaff[3], mf_unaff[4], fm_unaff[4], mf_unaff[5], fm_unaff[5]);

         fprintf(file, "%2s[%11s] %8d %8d %8d %8d %8d %8d\n", " ", "Discordant",
           mf_disc[3], fm_disc[3], mf_disc[4], fm_disc[4], mf_disc[5], fm_disc[5]);

         fprintf(file, "%2s[%11s] %8d %8d %8d %8d %8d %8d\n", " ", "Affected",
           mf_aff[3], fm_aff[3], mf_aff[4], fm_aff[4], mf_aff[5], fm_aff[5]);
         }
      }
   fprintf(file, "\n");
   }

void PrintDetailHeader(FILE * file, const char * additional)
   {
   fprintf(file, "\nOrdered Pair Detail %s:\n", additional ? additional : "");

   fprintf(file, "%15s %17s %17s %17s\n", " ",
           "ParentChild", "Grandparent", "Avuncular");

   fprintf(file, "%15s %8s %8s %8s %8s %8s %8s\n"," ",
           "F/D", "M/S", "GF/GD", "GM/GS", "U/N", "A/N");
   }

void PairTraitStatistics(FILE * file, PedigreePairs & pairs, bool is_covariate)
   {
   int filter_limit = (is_covariate ? Person::covariateCount : Person::traitCount);
   if (filter_limit == 0) return;

   IntArray * cts = new IntArray[filter_limit];
   StringArray * corr_txt = new StringArray[filter_limit];

   GetPairTraitStatistics(file, pairs, is_covariate, corr_txt, cts);

   String trait_name;

   fprintf(file, "\n\nPair Correlations for Each %s:\n", is_covariate ? "Covariate" : "Trait");
   fprintf(file, "%15s %8s %8s %8s %12s %12s %9s\n", " " ,
           "Sib", "HalfSib", "Cousin", "ParentChild", "Grandparent", "Avuncular");

   for(int i = 0; i < filter_limit; i++)
      {
      trait_name = is_covariate ? Person::covariateNames[i] : Person::traitNames[i] ;

      fprintf(file, "%15s %8s %8s %8s %12s %12s %9s\n",(const char *) trait_name,
             (const char *) corr_txt[i][0], (const char *) corr_txt[i][1],
             (const char *) corr_txt[i][2], (const char *) corr_txt[i][3],
             (const char *) corr_txt[i][4], (const char *) corr_txt[i][5]);
      }

   fprintf(file, "\nPair Counts for Each %s:\n", is_covariate ? "Covariate" : "Trait");
   fprintf(file, "%15s %8s %8s %8s %12s %12s %9s\n", " ",
           "Sib", "HalfSib", "Cousin", "ParentChild", "Grandparent", "Avuncular");

   for( int i = 0; i < filter_limit; i++)
      {
      trait_name = is_covariate ? Person::covariateNames[i] : Person::traitNames[i] ;

      fprintf(file, "%15s %8d %8d %8d %12d %12d %9d\n", (const char *) trait_name,
              cts[i][0], cts[i][1], cts[i][2], cts[i][3], cts[i][4], cts[i][5]);
      }

   if (pairs.sibs.gender == "Opposite" && (pairs.avuncularCt + pairs.grandParentCt + pairs.parentCt > 0))
      {
      PedigreePairs mf_pairs(ped), fm_pairs(ped);
      IntArray  *   mf_cts = new IntArray[filter_limit];
      IntArray  *   fm_cts = new IntArray[filter_limit];
      StringArray * mf_corr = new StringArray[filter_limit];
      StringArray * fm_corr = new StringArray[filter_limit];

      pairs.SplitOnGenderOrder(mf_pairs, fm_pairs);

      GetPairTraitStatistics(file, fm_pairs, is_covariate, fm_corr, fm_cts, false);
      GetPairTraitStatistics(file, mf_pairs, is_covariate, mf_corr, mf_cts, false);

      bool need_header = true;
      for( int i = 0; i < filter_limit; i++)
         {
         if (fm_cts[i][3] + fm_cts[i][4] + fm_cts[i][5]
             + mf_cts[i][3] + mf_cts[i][4] + mf_cts[i][5] <= 0) continue;

         if (need_header) PrintDetailHeader(file, "- Correlations");
         need_header = false;

         String trait_name = is_covariate ? Person::covariateNames[i] : Person::traitNames[i];
         fprintf(file, "%15s %8s %8s %8s %8s %8s %8s\n", (const char *) trait_name,
                 (const char *) mf_corr[i][3], (const char *) fm_corr[i][3],
                 (const char *) mf_corr[i][4], (const char *) fm_corr[i][4],
                 (const char *) mf_corr[i][5], (const char *) fm_corr[i][5]);
         }

      need_header = true;
      for( int i = 0; i < filter_limit; i++)
         {
         if (fm_cts[i][3] + fm_cts[i][4] + fm_cts[i][5]
             + mf_cts[i][3] + mf_cts[i][4] + mf_cts[i][5] <= 0) continue;

         if (need_header) PrintDetailHeader(file, "- Counts");
         need_header = false;

         String trait_name = is_covariate ? Person::covariateNames[i] : Person::traitNames[i];
         fprintf(file, "%15s %8d %8d %8d %8d %8d %8d\n", (const char *) trait_name,
                 mf_cts[i][3], fm_cts[i][3], mf_cts[i][4], fm_cts[i][4], mf_cts[i][5], fm_cts[i][5]);
         }

      delete [] mf_cts;
      delete [] fm_cts;
      delete [] mf_corr;
      delete [] fm_corr;
      }
   delete [] cts;
   delete [] corr_txt;
   }

void GetPairTraitStatistics(FILE * file, PedigreePairs & pairs, bool is_covariate,
StringArray * corr_txt, IntArray * cts, bool do_unordered)
   {
   int filter_limit = (is_covariate ? Person::covariateCount : Person::traitCount);

   for (int i = 0; i < filter_limit; i++)
      {
      cts[i].Dimension(6);
      cts[i].Zero();
      corr_txt[i].Dimension(6);
      }

   for (int filter = 0; filter < filter_limit; filter++)
      {
      Vector correlations;
      correlations.Dimension(6);

      pairs.GetCorrelations(filter, is_covariate, correlations, cts[filter], do_unordered);

      for (int j = 0; j < 6; j++)
         if (cts[filter][j] > 1)
            corr_txt[filter][j].printf("%7.4lf", correlations[j]);
         else
            corr_txt[filter][j].printf("%7s", "-");
      }
   }

void ByFamilyStatistics()
   {
   FILE * covar_file = NULL;
   FILE * trait_file = NULL;
   FILE * affect_file = NULL;
   FILE * marker_file = NULL;
   FILE * pair_file = NULL;

   if (Person::covariateCount > 0)
      {
      covar_file = fopen("pedstats.cov.fam", "wt");
      if (covar_file == NULL)
         error("Error opening file [%s]\n", "pedstats.cov.fam");
      else
         fprintf(covar_file, "\nCOVARIATE STATISTICS BY FAMILY FOR PEDIGREE FILE: %s\n",
                 (const char * ) pedName);
      }

   if (Person::traitCount > 0)
      {
      trait_file = fopen("pedstats.trait.fam", "wt");
      if (trait_file == NULL)
         error("Error opening file [%s]\n", "pedstats.trait.fam");
      else
         fprintf(trait_file, "\nTRAIT STATISTICS BY FAMILY FOR PEDIGREE FILE: %s\n",
                 (const char * ) pedName);
      }

   if (Person::affectionCount > 0)
      {
      affect_file = fopen("pedstats.aff.fam", "wt");
      if (affect_file == NULL)
         error("Error opening file [%s]\n", "pedstats.aff.fam");
      else
         fprintf(affect_file, "\nAFFECTION STATISTICS BY FAMILY FOR PEDIGREE FILE: %s\n",
                 (const char * ) pedName);
      }

   if (Person::markerCount > 0)
      {
      marker_file = fopen("pedstats.mark.fam", "wt");
      if (marker_file == NULL)
         error("Error opening file [%s]\n", "pedstats.mark.fam");
      else
         fprintf(marker_file, "\nMARKER STATISTICS BY FAMILY FOR PEDIGREE FILE: %s\n",
                          (const char * ) pedName);
      }

   if (pairs)
      {
      pair_file = fopen("pedstats.pair.fam", "wt");
      if (pair_file == NULL)
         error("Error opening file [%s]\n", "pedstats.pair.fam");
      else
         fprintf(pair_file, "\nPAIR STATISTICS BY FAMILY FOR PEDIGREE FILE: %s\n",
                        (const char * ) pedName);
      }

   for (int f = 0; f < ped.familyCount; f++)
      {
      FilteredPedigree curr_ped;
      curr_ped.pd = ped.pd;
      // this subroutine also builds up the filter
      ped.ExtractFamily(f, curr_ped);

      PedigreePairs curr_pairs(curr_ped);
      curr_pairs.Build();

      String added_label;
      added_label.printf(" FOR FAMILY %s", (const char *) ped.families[f]->famid);

      TraitStatistics(trait_file, curr_ped, curr_pairs.sibs, (const char *) added_label);
      CovarStatistics(covar_file, curr_ped, curr_pairs.sibs, (const char *) added_label);
      AffectionStatistics(affect_file, curr_ped, (const char *) added_label);
      MarkerStatistics(marker_file, curr_ped, (const char *) added_label);

      if (pairs)
         PairStatistics(pair_file, curr_ped, curr_pairs, (const char *) added_label);
      }

   if (trait_file)  fclose(trait_file);
   if (covar_file)  fclose(covar_file);
   if (affect_file) fclose(affect_file);
   if (marker_file) fclose(marker_file);
   if (pair_file)   fclose(pair_file);
   }

void CheckAges(PedigreePairs & pairs)
   {
   if (Person::traitCount == 0 && Person::covariateCount == 0) return;

   String current_trait;
   bool found_name = false;

   for (int j = 0; j < Person::traitCount; j++)
      {
      if (Person::traitNames[j] != ageLabel) continue;
      found_name = true;
      CheckAgeValues(pairs, j, false);
      }

   for (int j = 0; j < Person::covariateCount; j++)
      {
      if (Person::covariateNames[j] != ageLabel) continue;
      found_name = true;
      CheckAgeValues(pairs, j, true);
      }

   if (!found_name)
      error("An age ordering check was requested for the variable %s, but a variable\n"
            "with this name was not found in this data set. Age labels are case sensitive\n"
            "and must correspond exactly to that listed in your dat file", (const char *) ageLabel);
   }

void CheckAgeValues(PedigreePairs & pairs, int var, bool is_covariate)
   {
   bool exit = false;
   String ordered_trait = is_covariate ? Person::covariateNames[var] :
                                         Person::traitNames[var];

   printf("Checking age consistency among relative pairs for %s %s ... \n",
          is_covariate ? "covariate" : "trait", (const char *) ordered_trait);

   exit |= pairs.parents.CheckTraitOrdering(var, false, is_covariate, birthYear);
   exit |= pairs.grandparents.CheckTraitOrdering(var, false, is_covariate, birthYear);
   exit |= pairs.others.CheckTraitOrdering(var, true, is_covariate, birthYear);
   exit |= pairs.sibs.CheckSameTraitValue(var, true, is_covariate, birthYear);
   exit |= pairs.parents.CheckValueGap(var, shortGeneration, longGeneration, is_covariate, birthYear);
   exit |= pairs.grandparents.CheckValueGap(var, 2.0 * shortGeneration, 2.0 * longGeneration, is_covariate, birthYear);

   if (exit)
      error("Error in pairwise age values for %s %s",
            is_covariate ? "covariate" : "trait", (const char *) ordered_trait);
   }

void ByAffectionStatistics()
   {
   if (Person::affectionCount == 0) return;
   if (Person::covariateCount + Person::traitCount + Person::markerCount == 0) return;

   FILE * byaff_file = fopen("pedstats.byaff", "wt");
   if (byaff_file == NULL)
      error("Error opening file [%s]\n", "pedstats.byaff");
   else
      fprintf(byaff_file, "\nPEDIGREE STATISTICS BY AFFECTION STATUS FOR PEDIGREE FILE: %s\n",
                 (const char * ) pedName);

   String label;
   for(int i = 0; i < Person::affectionCount; i++)
      {
      label.printf(" FOR PERSONS AFFECTED BY %s", (const char * )Person::affectionNames[i]);
      GetByAffectionStatistics(i, byaff_file, (const char *) label, 2);

      label.printf(" FOR PERSONS UNAFFECTED BY %s", (const char *) Person::affectionNames[i]);
      GetByAffectionStatistics(i, byaff_file, (const char *) label, 1);
      }
   if (byaff_file) fclose(byaff_file);
   }

void GetByAffectionStatistics(int a, FILE * file, const char * label, int affection_target)
   {
   FilteredPedigree new_ped;
   new_ped.pd = ped.pd;
   ped.ExtractOnAffection(a, new_ped, affection_target);

   PedigreePairs new_pairs(new_ped);
   new_pairs.Build();

   TraitStatistics(file, new_ped, new_pairs.sibs, label);
   CovarStatistics(file, new_ped, new_pairs.sibs, label);
   MarkerStatistics(file, new_ped, label);

   if (pairs)
      PairStatistics(file, new_ped, new_pairs, label, true);
   }

void FilterPedigree(FilteredPedigree & ped)
   {
   if (minGenos > Person::markerCount)
      error("Minimum number of genotypes exceeds number of markers in this dataset");

   if (minCovariates > Person::covariateCount)
      error("Minimum number of measured covariates exceeds number of covariates in this dataset");

   if (minPhenos > Person::traitCount)
      error("Minimum number of trait phenotypes exceeds number of traits in this dataset");

   int target_affection = -1;
   if (affectedFor != "")
      {
      bool found_affection = false;
      target_affection = 0;

      while (found_affection == false && target_affection < Person::affectionCount)
         {
         found_affection = (affectedFor == Person::affectionNames[target_affection]);
         if (found_affection == false)
            target_affection++;
         }

      if (found_affection == false)
         error("Unable to filter pedigree on affection variable %s. This \n"
               "variable is not present in this data set. Affection variable\n"
               "names are case sensitive and must correspond exactly to that\n"
               "listed in the dat file\n");
      }

   int filters_applied = 0;

   StringArray filter_labels;
   filter_labels.Dimension(4);
   for (int i = 0; i < 4; i++)
      filter_labels[i] =  "";

   IntArray filter;
   filter.Dimension(ped.count);
   filter.Zero();

   int poss_filter = ped.count;
   int to_filter = 0;

   if (target_affection != -1)
      {
      to_filter = BuildAffectedFilter(target_affection, ped, filter);

      filter_labels[filters_applied].printf("affected by %s", (const char *) affectedFor);

      printf("\n\nFiltering sample to remove individuals not diagnosed or unaffected by \n%s...\n", (const char *) affectedFor);
      printf("%d individuals of %d removed from pedigree sample\n", to_filter, poss_filter);

      ped.Filter(filter);
      poss_filter -= to_filter;
      filters_applied ++;
      }

   if (minGenos > 0)
      {
      to_filter = BuildGenotypedFilter(ped, filter);

      filter_labels[filters_applied].printf("genotyped on at least %d marker%s",
                                           minGenos,  minGenos != 1 ? "s" : "");

      printf("\n\nFiltering sample to remove individuals genotyped on less than %d "
             "marker%s ...\n", minGenos, minGenos != 1 ? "s" : "");
      printf("%d individuals of %d removed from pedigree sample\n", to_filter, poss_filter);

      ped.Filter(filter);
      poss_filter -= to_filter;
      filters_applied ++;
      }

   if (minPhenos > 0)
      {
      to_filter = BuildPhenotypedFilter(ped, filter, false);

      filter_labels[filters_applied].printf("phenotyped for at least %d trait%s",
                   minPhenos, minPhenos != 1 ? "s" : "");

      printf("\n\nFiltering sample to remove individuals phenotyped on less than %d trait%s ...\n",
             minPhenos, minPhenos != 1 ? "s" : "");
      printf("%d individuals of %d removed from pedigree sample\n", to_filter, poss_filter);

      ped.Filter(filter);
      poss_filter -= to_filter;
      filters_applied ++;
      }

   if (minCovariates > 0)
      {
      to_filter = BuildPhenotypedFilter(ped, filter, true);

      filter_labels[filters_applied].printf("controlled for at least %d covariate%s",
                 minCovariates, minCovariates != 1 ? "s" : "");

      printf("\n\nFiltering sample to remove individuals controlled on less than"
             " %d covariate%s ...\n", minCovariates, minCovariates != 1 ? "s" : "");
      printf("%d individuals of %d removed from pedigree sample\n", to_filter, poss_filter);

      ped.Filter(filter);
      poss_filter -= to_filter;
      filters_applied ++;
      }

   StringArray switches;
   switches.Dimension(3);
   for (int i = 0; i < 3; i++)
      switches[i] = "";

   if (filters_applied > 1)
      {
      switches[filters_applied - 2] = " and ";
      if (filters_applied > 2)
         {
         switches[filters_applied - 3] = ", ";
         if (filters_applied > 3)
            switches[filters_applied - 4] = ", ";
         }
      }

   textFilterLabel.printf("individuals %s%s\n%s%s%s%s\n%s%s%s",
                         (const char *) filter_labels[0],
                         (const char *) switches[0], "  ",
                         (const char *) filter_labels[1],  (const char *) switches[1],
                         (const char *) filter_labels[2],
                         switches[2] == " and " ? " " : "  ", (const char *) switches[2],
                         (const char *) filter_labels[3]);

   pdfFilterLabel.printf("* NOTE:  Sample restricted to individuals %s%s%s%s%s%s%s",
                         (const char *) filter_labels[0], (const char *) switches[0],
                         (const char *) filter_labels[1], (const char *) switches[1],
                         (const char *) filter_labels[2], (const char *) switches[2],
                         (const char *) filter_labels[3]);
   }

int BuildGenotypedFilter(FilteredPedigree & ped, IntArray & filter)
   {
   int to_filter = 0;
   filter.Zero();

   for (int i = 0; i < ped.count; i++)
      {
      if (ped[i].ngeno < minGenos)
        filter[i] = 1;
      if (ped.filtered[i] == 0 && filter[i] == 1)
        to_filter++;
      }
   return to_filter;
   }

int BuildPhenotypedFilter(FilteredPedigree & ped, IntArray & filter, bool is_covariate)
   {
   int to_filter = 0;
   filter.Zero();

   int count = 0;
   int min_values = is_covariate ? minCovariates : minPhenos;

   for (int i = 0; i < ped.count; i++)
      {
      count = 0;

      if (is_covariate)
         for (int c = 0; c < Person::covariateCount; c++)
            count += (ped[i].isControlled(c));
      else
         for (int t = 0; t < Person::traitCount; t++)
            count += (ped[i].isPhenotyped(t));

      if (count < min_values)
            filter[i] = 1;

      if (ped.filtered[i] == 0 && filter[i] == 1)
         to_filter++;
      }
   return to_filter;
   }

int BuildAffectedFilter(int target_affection, FilteredPedigree & ped, IntArray & filter)
   {
   int to_filter = 0;
   filter.Zero();

   for (int i = 0; i < ped.count; i++)
      {
      if (ped[i].affections[target_affection] != 2)
         filter[i] = 1;
      if (ped.filtered[i] == 0 && filter[i] == 1)
         to_filter++;
      }
   return to_filter;
   }
 
